from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "custom_icons.zip",down_path)
download_file(xz_url + "library.zip",down_path)


bl_info = {
    "name": "ProFlow",
    "description": "",
    "author": "Cedric Lepiller aka Pitiwazou",
    "version": (0, 1, 4),
    "blender": (4, 1, 0),
    "location": "Property Panel, Press N in the 3DView",
    "wiki_url": "",
    "category": "TOOLS"}


# Import des modules
if "bpy" in locals():
    import importlib
    reloadable_modules = [
        "ui",
        "operators",
        "preview",
        "modals"
    ]
    for module in reloadable_modules:
        if module in locals():
            importlib.reload(locals()[module])

from . import  (ui,
                operators,
                preview,
                modals
                )

import bpy
from bpy.types import Operator, Menu, Panel, PropertyGroup, AddonPreferences, Scene, WindowManager, BlendData
from bpy.props import (StringProperty,
                       BoolProperty,
                       PointerProperty,
                       FloatVectorProperty,
                       FloatProperty,
                       EnumProperty,
                       IntProperty,
                       BoolVectorProperty,
                       CollectionProperty)

from .ui import *
import rna_keymap_ui
from .custom_icons.icons import load_icons
from bpy.types import WindowManager
from .functions import get_addon_preferences, set_custom_mesh_from_selection


keymaps_items_dict = {"PROFLOW POPUP":['proflow.popup', None, '3D View '
                                      'Generic', 'VIEW_3D', 'WINDOW',
                                      'Q', 'PRESS', True, True, False
                                      ],

                      "PROFLOW EDIT MODAL":['object.proflow_modal_rotate_scale', None, '3D View '
                                      'Generic', 'VIEW_3D', 'WINDOW',
                                      'Q', 'PRESS', True, True, True
                                      ],

                     }

def update_proflow_category(self, context):
    is_panel = hasattr(bpy.types, 'PROFLOW_PT_panel')

    if is_panel:
        try:
            bpy.utils.unregister_class(PROFLOW_PT_panel)
        except:
            pass
    # PROFLOW_PT_panel.bl_category = self.category
    bpy.utils.register_class(PROFLOW_PT_panel)

# -----------------------------------------------------
# LIBRARY
# -----------------------------------------------------
class PROFLOW_PropertyGroup_categories(bpy.types.PropertyGroup):
    path : StringProperty(subtype = 'DIR_PATH')

class PROFLOW_OT_add_categories_path(Operator):
    bl_idname = "object.add_categories_path"
    bl_label = "Add Categories Path"
    bl_description = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        addon_pref = get_addon_preferences()
        addon_pref.categories.add()
        return {'FINISHED'}

class PROFLOW_OT_remove_categories_path(Operator):
    bl_idname = "object.remove_categories_path"
    bl_label = "Add Categories Path"
    bl_description = ""
    bl_options = {'REGISTER', 'UNDO'}

    index : IntProperty()

    def execute(self, context):
        addon_pref = get_addon_preferences()

        addon_pref.categories.remove(self.index)
        return {'FINISHED'}

def call_back_enum(self, context):
    addon_pref = get_addon_preferences()
    directory = os.path.join(os.path.dirname(__file__), "library")
    mes_reps = ["welds", "customs", "profiles"]
    categories = [(os.path.join(directory, rep), rep, "" ) for rep in mes_reps]

    for cat in addon_pref.categories:
        root, dirs, files = next(os.walk(cat.path)) # on liste le path, reps et noms de fichiers
        if "files" in dirs and "icons" in dirs:
            categories.append((cat.path, os.path.basename(os.path.abspath(cat.path)), ""))

    return categories

# -----------------------------------------------------
# PREFERENCES
# -----------------------------------------------------
class PROFLOW_MT_addon_preferences(AddonPreferences):
    bl_idname = __package__

    prefs_tabs: EnumProperty(
        items=(('options', "Options", "Options"),
               ('keymaps', "Keymaps", "Keymaps"),),
        default='options')

    category: StringProperty(description="Choose a name for the category of the panel",
                             default="ProFlow",
                             update=update_proflow_category)

    categories: CollectionProperty(type=PROFLOW_PropertyGroup_categories) # prefs.categories.index
    show_path_info: BoolProperty(default=False)
    preview_scale: IntProperty(name="", default=7, min=3, max=10)
    preview_scale_popup: IntProperty(name="", default=5, min=2, max=10)
    preview_show_labels: BoolProperty(name='', default=True)

    # --TEXT OPTIONS
    text_color: FloatVectorProperty(name="", default=(1, 1, 1, 1), min=0, max=1, size=4, subtype='COLOR_GAMMA')
    text_color_1: FloatVectorProperty(name="", default=(0.5, 1, 0, 1), min=0, max=1, size=4, subtype='COLOR_GAMMA')
    text_color_2: FloatVectorProperty(name="", default=(0, 0.7, 1, 1), min=0, max=1, size=4, subtype='COLOR_GAMMA')
    text_size: IntProperty(name="",default=15, min=10, max=50)
    text_shadow: BoolProperty(default=True)
    text_pos_y: IntProperty(name="",default=80,min=0, max=2000,description="Position of the text in Y")
    text_pos_x: IntProperty(name="",default=2,min=0, max=2000,description="Position of the text in X")
    text_color: FloatVectorProperty(name="",default=(1, 1, 1, 1),min=0, max=1, size=4,subtype='COLOR_GAMMA')
    text_color_1: FloatVectorProperty(name="",default=(0.5, 1, 0, 1),min=0, max=1, size=4,subtype='COLOR_GAMMA')
    text_color_2: FloatVectorProperty(name="",default=(0, 0.7, 1, 1),min=0, max=1, size=4,subtype='COLOR_GAMMA')
    text_space: FloatProperty(name="Space Between lines",default=2,min=0.5, max=100,description="Space Between lines")


    add_to_collection: BoolProperty(name='ADD TO COLLECTION', default=True, description='Add Flows to a Collection')
    # rotate_on_modal: BoolProperty(name='', default=True)
    copy_material: BoolProperty(name='COPY MATERIAL', default=True,
                                description='Copy Material from the selected Object to the new created Flows')

    modal_speed: FloatProperty(name="", default=1.0, min=0.001, max=1000,
                                     description="Speed to change the Focus of the Camera")

    edit_overlay: BoolProperty(name='EDIT OVERLAY', default=True,
                                description='Show wireframe Overlay when necessary (Cut and Extract tool)')

    proflow_simplify: BoolProperty(name='SHOW HIDE MODIFIERS', default=True,description='Show Hide Modifiers')

    info: BoolProperty(
        name="",
        default=False,
        description="INFO")

    def draw(self, context):
        layout = self.layout
        icons = load_icons()
        row = layout.row(align=True)
        row.prop(self, "prefs_tabs", expand=True)

        if self.prefs_tabs == 'options':

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.label(text="PANEL CATEGORY")
            row.prop(self, "category", text="")

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.label(text="OPTIONS")
            row = col.row(align=True)
            row.label(text="Add to Collection")
            row.prop(self, "add_to_collection", text="      ")
            row = col.row(align=True)
            row.label(text="Copy Material")
            row.prop(self, "copy_material", text="      ")
            row = col.row(align=True)
            row.label(text="Edit Overlay")
            row.prop(self, "edit_overlay", text="      ")

            row = box.row(align=True)
            row.label(text="Modal Speed")
            row.prop(self, "modal_speed")

            # box = layout.box()
            # split = box.split()
            # col = split.column()
            # row = col.row(align=True)
            # row.label(text="MODAL OPTIONS")
            # row = col.row(align=True)
            # row.label(text="Auto Rotate in the Modal")
            # row.prop(self, "rotate_on_modal", text="      ")

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.label(text="MODAL TEXTS SETTINGS")
            row = box.row(align=True)
            row.label(text="Text Size")
            row.prop(self, "text_size")

            row = box.row(align=True)
            row.label(text="Text X position")
            row.prop(self, "text_pos_x")

            row = box.row(align=True)
            row.label(text="Text Y position")
            row.prop(self, "text_pos_y")

            # row = box.row(align=True)
            # row.label(text="Title Color")
            # row.prop(self, "text_color")
            #
            # row = box.row(align=True)
            # row.label(text="Settings Color")
            # row.prop(self, "text_color_1")
            #
            # row = box.row(align=True)
            # row.label(text="Value Color")
            # row.prop(self, "text_color_2")

            row = box.row(align=True)
            row.label(text="Activate Shadows")
            row.prop(self, "text_shadow", text="      ")

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.label(text="ASSET LIBRARY SETTINGS")
            row = col.row(align=True)
            row.label(text="Preview Thumbnails Scale")
            row.prop(self, "preview_scale")
            row = col.row(align=True)
            row.label(text="Popup Thumbnails Scale")
            row.prop(self, "preview_scale_popup")
            row = col.row(align=True)
            row.label(text="Show thumbnails Labels")
            row.prop(self, "preview_show_labels", text="      ")

            for index, cat in enumerate(self.categories):
                if self.show_path_info:
                    box = layout.box()
                    split = box.split()
                    col = split.column()

                row = col.row(align=True)
                row.prop(cat, 'path', text="")
                row.operator('object.remove_categories_path', text="", icon='REMOVE').index = index

        # -----------Keymap settings
        if self.prefs_tabs == 'keymaps':
            wm = bpy.context.window_manager

            draw_keymap_items(wm, layout)

# -----------------------------------------------------------------------------
#    Keymap
# -----------------------------------------------------------------------------
addon_keymaps = []

def draw_keymap_items(wm, layout):
    kc = wm.keyconfigs.user

    for name, items in keymaps_items_dict.items():
        kmi_name, kmi_value, km_name = items[:3]
        box = layout.box()
        split = box.split()
        col = split.column()
        col.label(text=name)
        col.separator()
        km = kc.keymaps[km_name]
        get_hotkey_entry_item(kc, km, kmi_name, kmi_value, col)


def get_hotkey_entry_item(kc, km, kmi_name, kmi_value, col):

    # for menus and pie_menu
    if kmi_value:
        for km_item in km.keymap_items:
            if km_item.idname == kmi_name and km_item.properties.name == kmi_value:
                col.context_pointer_set('keymap', km)
                rna_keymap_ui.draw_kmi([], kc, km, km_item, col, 0)
                return

        col.label(text = "No hotkey entry found for {}".format(kmi_value))
        col.operator(PROFLOW_OT_add_hotkey.bl_idname, icon='ZOOMIN')

    # for operators
    else:
        if km.keymap_items.get(kmi_name):
            col.context_pointer_set('keymap', km)
            rna_keymap_ui.draw_kmi(
                    [], kc, km, km.keymap_items[kmi_name], col, 0)
        else:
            col.label(text = "No hotkey entry found for {}".format(kmi_name))
            col.operator(PROFLOW_OT_add_hotkey.bl_idname, icon='ZOOMIN')


class PROFLOW_OT_add_hotkey(bpy.types.Operator):
    ''' Add hotkey entry '''
    bl_idname = "template.add_hotkey"
    bl_label = "Add Hotkeys"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        add_hotkey()

        self.report({'INFO'},
                    "Hotkey added in User Preferences -> Input -> Screen -> Screen (Global)")
        return {'FINISHED'}

def add_hotkey():
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon

    if not kc:
        return

    for items in keymaps_items_dict.values():
        kmi_name, kmi_value, km_name, space_type, region_type = items[:5]
        eventType, eventValue, ctrl, shift, alt = items[5:]
        km = kc.keymaps.new(name = km_name, space_type = space_type,
                            region_type=region_type)

        kmi = km.keymap_items.new(kmi_name, eventType,
                                  eventValue, ctrl = ctrl, shift = shift,
                                  alt = alt

                                  )
        if kmi_value:
            kmi.properties.name = kmi_value

        kmi.active = True

    addon_keymaps.append((km, kmi))


def remove_hotkey():
    ''' clears all addon level keymap hotkeys stored in addon_keymaps '''

    kmi_values = [item[1] for item in keymaps_items_dict.values() if item]
    kmi_names = [item[0] for item in keymaps_items_dict.values() if item not in ['wm.call_menu', 'wm.call_menu_pie']]

    for km, kmi in addon_keymaps:
        # remove addon keymap for menu and pie menu
        if hasattr(kmi.properties, 'name'):
            if kmi_values:
                if kmi.properties.name in kmi_values:
                    km.keymap_items.remove(kmi)

        # remove addon_keymap for operators
        else:
            if kmi_names:
                if kmi.name in kmi_names:
                    km.keymap_items.remove(kmi)

    addon_keymaps.clear()

# -----------------------------------------------------
# PROPERTYGROUP
# -----------------------------------------------------
class PROFLOW_PropertyGroup(bpy.types.PropertyGroup):
    custom_mesh : PointerProperty(type=bpy.types.Object)
    active_mesh: PointerProperty(type=bpy.types.Object)
    target_mesh: PointerProperty(type=bpy.types.Object)
    auxiliary_mesh: PointerProperty(type=bpy.types.Object)
    replace_asset_curve: PointerProperty(type=bpy.types.Object)
    replace_asset_mesh: PointerProperty(type=bpy.types.Object)
    cyclic : BoolProperty(name="", description="Make the curve Cyclic (Closed)", default=True)
    use_custom_mesh : BoolProperty(name='CUSTOM ASSET', default=False,
                                   description='Use Custom Asset instead of the meshes in the preview',
                                   update=set_custom_mesh_from_selection)
    show_modifiers : BoolProperty(name='', default=False, description='Show Assets Modifiers and edit it')
    show_curves_settings: BoolProperty(name='', default=False, description='Show Curve Settings')
    no_material : BoolProperty(name='', default=False, description='Object has no material')
    act_obj_material_name : StringProperty(description='')
    replace_asset : BoolProperty(name='REPLACE ASSET', default=False, description='Replace the current asset')
    is_set_mat : BoolProperty(name='', default=False)
    is_set_replace_asset: BoolProperty(name='', default=False)
    ui_categories: EnumProperty(name="", items=call_back_enum)
    parent_asset : BoolProperty(name='Parent Asset to Active Object', default=False)
    use_plane: BoolProperty(name='Parent Asset to a plane to avoid deformations', default=False)
    draw: BoolProperty(name='Draw the Curve', default=False)
    use_selected_custom_mesh: BoolProperty(name='Use Selected Mesh to Draw', default=False)
    draw_on_surface: BoolProperty(name='DRAW ON SURFACE', default=True, description='If Disabled >>> draw on Cursor')
    edit_with_modal: BoolProperty(name='EDIT WITH MODAL', default=True, description='Edit the resulting Asset with Modal to rotate, scale, etc')
    edit_axis: BoolProperty(name='', default=True,description='')
    use_plane_instance: BoolProperty(name='', default=False, description='Add a Plane to avoid Deformations on the asset')
    from_booleans: BoolProperty(default=False)
    show_options: BoolProperty(default=False)
    Transfert_normals: BoolProperty(name='TRANSFERT NORMALS',default=False, description='Transfert Normals from the Active Object to the Asset Object')
    random_location: BoolProperty(name='RANDOM LOCATION ON CONVERTION', default=False,
                                    description='Make a random location when converting to MESH')
    random_rotation: BoolProperty(name='RANDOM ROTATION ON CONVERTION', default=False,
                                  description='Make a random rotation when converting to MESH')
    random_scale: BoolProperty(name='RANDOM SCALE ON CONVERTION', default=False,
                                  description='Make a random Scale when converting to MESH')
    join_meshes: BoolProperty(name='JOIN MESHES ON CONVERTION', default=False,
                                  description='Join objects when converting to MESH')

    link_objects: BoolProperty(name='LINKS OBJECTS AFTER CONVERTION', default=False,
                                  description='Link objects after converting to MESH')

    overlay_state: BoolProperty(name='JOIN MESHES ON CONVERTION', default=False,
                                  description='Join objects when converting to MESH')

    overlay_wire_state: BoolProperty(name='JOIN MESHES ON CONVERTION', default=False,
                                  description='Join objects when converting to MESH')

    remove_subsurf: BoolProperty(name='REMOVE SUBSURF MODIFIER', default=False,
                                  description='Remove the subsurf Modifier when converting the objects')

    random_transforms: BoolProperty(name='RANDOM TRANSFORM MODIFIER', default=False,
                                 description='Launch Random Transforms Modal after converting the Objects')

    keep_wireframe: BoolProperty(name='KEEP WIREFRAME VISIBLE IN MODAL', default=True,
                                 description='Keep the wireframe visible when you are in the modal')

    from_modals: BoolProperty(name='', default=False,
                                 description='')

    active_coll: StringProperty()


    # current_mesh_matrix_save = None

    directions: EnumProperty(
        items=(('x', "X", ""),
               ('y', "Y", ""),
               ('z', "Z", ""),
               ),
        default='z',
        update=axis_change,
        description='Directions (X,Y or Z)')


# -----------------------------------------------------
# REGISTER/UNREGISTER
# -----------------------------------------------------
CLASSES =  [PROFLOW_OT_add_categories_path,
            PROFLOW_PropertyGroup_categories,
            PROFLOW_OT_remove_categories_path,
            PROFLOW_MT_addon_preferences,
            PROFLOW_PropertyGroup,
            ]

# Register
def register():
    ui.register()
    operators.register()
    preview.register()
    modals.register()

    for cls in CLASSES:
        try:
            bpy.utils.register_class(cls)
        except:
            print(f"{cls.__name__} already registred")

    # PropertyGroup
    bpy.types.WindowManager.proflow = PointerProperty(type=PROFLOW_PropertyGroup)

    # Icons
    bpy.types.WindowManager.proflow_previews = EnumProperty(name="Thumbnails",
                                                            items=generate_previews,
                                                            description="Select a Flow")

    # Library
    pcoll = bpy.utils.previews.new()
    pcoll.images_location = ""
    pcoll.proflow_previews = ()
    proflow_collection["main"] = pcoll

    # Update Category
    context = bpy.context
    prefs = context.preferences.addons[__name__].preferences
    update_proflow_category(prefs, context)

    # hotkey setup
    add_hotkey()

def unregister():
    ui.unregister()
    operators.unregister()
    preview.unregister()
    modals.unregister()

    for previews in proflow_collection.values():
        bpy.utils.previews.remove(previews)

    proflow_collection.clear()

    for cls in reversed(CLASSES):
        try:
            bpy.utils.unregister_class(cls)
        except RuntimeError:
            pass
    # for cls in CLASSES:
    #     if hasattr(bpy.types, cls.__name__):
    #         bpy.utils.unregister_class(cls)

    # hotkey cleanup
    remove_hotkey()


if __name__ == "__main__":
    register()
