# -*- coding:utf-8 -*-

# ProFlow Add-on
# Copyright (C) 2018 Cedric Lepiller aka Pitiwazou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# <pep8 compliant>


import bpy
from bpy.types import Operator
from mathutils import Matrix, Vector
import bmesh
import math
from math import pi, radians, degrees
import blf
from bpy.props import (StringProperty,
                       IntProperty,
                       FloatProperty,
                       BoolProperty,
                       FloatVectorProperty,
                       EnumProperty
                       )
import random
from random import uniform
from .functions import get_addon_preferences
import numpy as np

class MyHandleClass:
    _handler = None

    @classmethod
    def add_handler(cls, function, args=()):
        cls._handler = bpy.types.SpaceView3D.draw_handler_add(
                function, args, 'WINDOW', 'POST_PIXEL'
                )

    @classmethod
    def remove_handler(cls):
        if cls._handler is not None:
            bpy.types.SpaceView3D.draw_handler_remove(cls._handler, 'WINDOW')
            cls._handler = None

# ----------------------------------------------------------------------------------------------------------------------
# DRAW TEXT ------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------
font_info = {
    "font_id": 0,
    "handler": None,
}

def proflow_draw_callback_px(x, y, packed_strings, obj):
    addon_pref = get_addon_preferences()

    text_size = addon_pref.text_size
    font_id = 0
    dpi = 1
    if bpy.app.version < (4, 0, 0):
        blf.size(font_id, text_size, 72)
    else:
        blf.size(font_id, text_size * dpi)
    # blf.size(font_id, text_size, 72)
    x_offset = 0
    y_offset = 0
    shadow_color = (0, 0, 0, 0.2)
    shadow_x = 2
    shadow_y = -2
    text_pos_x = addon_pref.text_pos_x
    text_pos_y = addon_pref.text_pos_y
    overlap = bpy.context.preferences.system.use_region_overlap
    t_panel_width = 0
    text_shadow = addon_pref.text_shadow
    if overlap:
        for region in bpy.context.area.regions:
            if region.type == 'TOOLS':
                t_panel_width = region.width

    width = bpy.context.region.width
    height = bpy.context.region.height

    pos_x = min((width - t_panel_width - 250) * text_pos_x / 100 + t_panel_width, width - 250)
    pos_y = height / 3 + text_pos_y

    line_height = (blf.dimensions(font_id, "M")[1] )

    for command in packed_strings:
        if len(command) == 2:
            pstr, pcol = command
            blf.color(0, *pcol)
            text_width, text_height = blf.dimensions(font_id, pstr)
            if text_shadow:
                blf.enable(0, blf.SHADOW)
                blf.shadow_offset(0, shadow_x, shadow_y)
                blf.shadow(0, 3, shadow_color[0], shadow_color[1], shadow_color[2], shadow_color[3])
            blf.position(font_id, (pos_x + x_offset), (pos_y + y_offset), 0)
            blf.draw(font_id, pstr)
            x_offset += text_width
        else:
            x_offset = 0
            y_offset -= line_height

x = 0
y = 0
y1 = 100
RED = (1, 0, 0, 1)
GREEN = (0, 1, 0, 1)
BLUE = (0, 0, 1, 1)
WHITE = (1,1,1,1)

CR = "Carriage Return"
ps = [("MODAL CUT", WHITE),CR,
      ("", WHITE),CR,
      ("", WHITE),CR,
      ("Click to Start Cutting", GREEN),CR,
      ("", WHITE),CR,
      ("", WHITE),CR,
      ("C: ", WHITE), ("Angle", GREEN),CR,
      ("", WHITE),CR,
      ("Z: ", WHITE), ("Cut Through", GREEN),CR,
      ("", WHITE),CR,
      ("CTRL+SPACE: ", WHITE), ("Extract Edge", GREEN),CR,
      ("", WHITE),CR,
      ("SPACE/ENTER: ", WHITE), ("Valid", GREEN),CR,
      ("", WHITE),CR,
      ("ESC: ", WHITE), ("Escape", GREEN)
      ]

ps1 = [("MODAL EXTRACT", WHITE),CR,
      ("", WHITE),CR,
      ("", WHITE),CR,
      ("Select Some Edges", GREEN),CR,
      ("", WHITE),CR,
      ("", WHITE),CR,
      ("CTRL + R: ", WHITE), ("Add Loop", GREEN),CR,
      ("", WHITE),CR,
      ("2: ", WHITE), ("Edges", GREEN),CR,
      ("", WHITE),CR,
      ("3: ", WHITE), ("Faces", GREEN),CR,
      ("", WHITE),CR,
      ("I: ", WHITE), ("Insert", GREEN),CR,
      ("", WHITE),CR,
      ("CTRL+SPACE: ", WHITE), ("Extract Edge", GREEN),CR,
      ("", WHITE),CR,
      ("SPACE/ENTER: ", WHITE), ("Valid", GREEN),CR,
      ("", WHITE),CR,
      ("ESC: ", WHITE), ("Escape", GREEN)
      ]
ps2 = [
      ("", WHITE),CR,
      ("MODAL EDIT FLOWS", WHITE),CR,
      ("", WHITE),CR,
      ("R: ", WHITE), ("Rotate", GREEN),CR,
      ("", WHITE),CR,
      ("S: ", WHITE), ("Scale", GREEN),CR,
      ("", WHITE),CR,
      ("Q: ", WHITE), ("Array Space", GREEN),CR,
      ("", WHITE),CR,
      ("C: ", WHITE), ("Cyclic", GREEN),CR,
      ("", WHITE),CR,
      ("F: ", WHITE), ("Flip Curve", GREEN),CR,
      ("", WHITE),CR,
      ("D: ", WHITE), ("Displace 1", GREEN),(" (if supported)", WHITE),CR,
      ("", WHITE),CR,
      ("G: ", WHITE), ("Displace 2", GREEN),(" (if supported)", WHITE),CR,
      ("", WHITE),CR,
      ("B: ", WHITE), ("Bend", GREEN),(" (if supported)", WHITE),CR,
      ("", WHITE),CR,
      ("P: ", WHITE), ("Poly/Bezier", GREEN),CR,
      ("", WHITE),CR,
      ("X: ", WHITE), ("Direction X", GREEN),CR,
      ("", WHITE),CR,
      ("Y: ", WHITE), ("Direction Y", GREEN),CR,
      ("", WHITE),CR,
      ("Z: ", WHITE), ("Direction Z", GREEN),CR,
      ("", WHITE),CR,
      ("CTRL+A: ", WHITE), ("Convert", GREEN),CR,
      ("", WHITE),CR,
      ("SPACE/ENTER: ", WHITE), ("Valid", GREEN),CR,
      ("", WHITE),CR,
      ("ESC: ", WHITE), ("Escape", GREEN)
      ]

ps3 = [("MODAL EDIT FLOATERS", WHITE),CR,
      # ("", WHITE),CR,
      # ("R: ", WHITE), ("Rotate", GREEN),CR,
      ("", WHITE),CR,
      ("S: ", WHITE), ("Scale Profile", GREEN),CR,
      ("", WHITE),CR,
      ("F: ", WHITE), ("Flip Curve Direction", GREEN),CR,
      ("", WHITE),CR,
      ("E: ", WHITE), ("Curve Offset", GREEN),CR,
      ("", WHITE),CR,
      ("D: ", WHITE), ("Resolution U", GREEN),CR,
      ("", WHITE),CR,
      ("G: ", WHITE), ("Resolution V", GREEN),CR,
      ("", WHITE),CR,
      ("H: ", WHITE), ("Fill Cap", GREEN),CR,
      ("", WHITE),CR,
      ("P: ", WHITE), ("Poly/Bezier", GREEN),CR,
       ]

ps4 = [("RANDOM TRANSFORMS", WHITE),CR,
      ("", WHITE),CR,
      ("R: ", WHITE), ("Rotate", GREEN),CR,
      ("", WHITE),CR,
      ("S: ", WHITE), ("Scale", GREEN),CR,
      ("", WHITE),CR,
      ("G: ", WHITE), ("Location", GREEN),CR,
      ("", WHITE),CR,
      ("Z: ", WHITE), ("Location Z", GREEN),CR,
      ("", WHITE),CR,
      ("SPACE/ENTER: ", WHITE), ("Valid", GREEN),CR,
      ("", WHITE),CR,
      ("ESC: ", WHITE), ("Undo", GREEN),CR,
      ("", WHITE),CR,
      ("RMB: ", WHITE), ("Undo and Exit", GREEN)
      ]
# draw_string(x, y, ps)



# def proflow_draw_callback_px(self, context):
#     x_offset = 0
#     y_offset = 0
#     overlap = context.preferences.system.use_region_overlap
#     t_panel_width = 0
#     if overlap:
#         for region in bpy.context.area.regions:
#             if region.type == 'TOOLS':
#                 t_panel_width = region.width
#
#
#     proflow_dpi = context.preferences.view.ui_scale
#     proflow_text_pos_x = context.area.width / 3 #* proflow_dpi
#
#     proflow_text_pos_y = context.area.height * 0.9 #+ (context.area.height * 0.25)
#
#     # x = t_panel_width + proflow_text_pos_x
#     x = proflow_text_pos_x
#     y = context.region.height - proflow_text_pos_y
#
#     font_id = font_info["font_id"]
#     Text =  "MODAL ON --- SPACE/ENTER: Valid, ESC: Escape"
#     text_width, text_height = blf.dimensions(font_id, Text)
#     blf.color(font_id, 1.0, 1.0, 1.0, 1.0)
#     blf.size(font_id, 50, 50)
#     blf.position(font_id, (x + x_offset), (y + y_offset), 0)
#     blf.draw(font_id, Text)
#     x_offset += text_width



class PROFLOW_OT_modal_cut(Operator):
    """
    FLOATERS FROM CUT

    Cut the mesh to create the Floaters.

    SPACE - Validate
    CTRL + SPACE - Extract Edge only
    """
    bl_idname = "object.proflow_modal_cut"
    bl_label = "ProFlow Modal"

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        if len([obj for obj in context.selected_objects if context.object is not None if obj.type == 'MESH']) == 1:
            return True

    # MODAL
    def modal(self, context, event):
        proflow = context.window_manager.proflow
        addon_pref = get_addon_preferences()

        if context.area is None:
            return {'FINISHED'}
        context.area.tag_redraw()

        if event.type in {'NUMPAD_1', 'NUMPAD_3', 'NUMPAD_7', 'NUMPAD_5'} or event.ctrl and event.type in {'NUMPAD_1',
                                                                                                           'NUMPAD_3',
                                                                                                           'NUMPAD_7',
                                                                                                           'NUMPAD_5'}:
            return {'PASS_THROUGH'}

        if event.type == 'MOUSEMOVE' and not self.start:
            self.start = True
            bpy.ops.mesh.knife_tool('INVOKE_DEFAULT', use_occlude_geometry=False, only_selected=False)

        if event.type in {'CTRL', 'SHIFT', 'C'}:
            return {'PASS_THROUGH'}

        if event.type == 'MIDDLEMOUSE' or event.alt and event.type == 'MIDDLEMOUSE':
            return {'PASS_THROUGH'}

        if event.type == 'LEFTMOUSE' and event.alt or event.type == 'RIGHTMOUSE' and event.alt:
            return {'PASS_THROUGH'}

        if event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} or event.ctrl and event.type in {'WHEELUPMOUSE',
                                                                                             'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        # Stop and finalize
        elif event.ctrl and event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'}:
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = False
            else:
                context.space_data.overlay.show_retopology = False
            context.object.show_in_front = False

            MyHandleClass.remove_handler()
            context.area.tag_redraw()

            me = context.object.data
            bm = bmesh.from_edit_mesh(me)
            if any([e.select for e in bm.faces]):
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.delete(type='FACE')
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.dissolve_mode()
                bpy.ops.mesh.delete(type='ONLY_FACE')
                bpy.ops.object.mode_set(mode='OBJECT')

                if proflow.is_set_replace_asset:
                    proflow.replace_asset = True
                if proflow.is_set_mat:
                    addon_pref.copy_material = True
                if addon_pref.edit_overlay:
                    # restore overlay
                    context.space_data.overlay.show_overlays = True
                    context.space_data.overlay.show_wireframes = True
                return {'FINISHED'}

            elif any([e.select for e in bm.edges]):
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.delete(type='VERT')
                bpy.ops.object.mode_set(mode='OBJECT')

                if proflow.is_set_replace_asset:
                    proflow.replace_asset = True
                if proflow.is_set_mat:
                    addon_pref.copy_material = True
                if addon_pref.edit_overlay:
                    # restore overlay
                    context.space_data.overlay.show_overlays = True
                    context.space_data.overlay.show_wireframes = True
                return {'FINISHED'}
            else:
                self.report({'WARNING'}, "Select some edges")



        # Stop and finalize
        elif event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'}:
            MyHandleClass.remove_handler()
            context.area.tag_redraw()

            bpy.ops.mesh.select_all(action='INVERT')
            bpy.ops.mesh.delete(type='VERT')
            bpy.ops.object.mode_set(mode='OBJECT')
            self.mesh_to_cut = context.active_object
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = False
            else:
                context.space_data.overlay.show_retopology = False
            # bpy.context.space_data.overlay.show_occlude_wire = False
            bpy.ops.object.convert(target='CURVE')

            bpy.ops.object.proflow_finalize()

            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True

            proflow.from_modals = True
            if addon_pref.edit_overlay:
                # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state
            return {'FINISHED'}

        # CANCEL AND EXIT
        elif event.type in {'ESC', 'RIGHTMOUSE'} and event.value == 'PRESS':
            MyHandleClass.remove_handler()
            context.area.tag_redraw()
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = False
            else:
                context.space_data.overlay.show_retopology = False
            # bpy.context.space_data.overlay.show_occlude_wire = False
            bpy.context.object.show_in_front = False

            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.delete(use_global=False)

            context.view_layer.objects.active = self.act_obj
            self.act_obj.select_set(state=True)

            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True
            if addon_pref.edit_overlay:
                # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):

        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "INFO: Do not work in Local Mode")
        #             return {'FINISHED'}

        if context.area.type == 'VIEW_3D' and context.object:
            addon_pref = get_addon_preferences()
            proflow = context.window_manager.proflow

            if addon_pref.edit_overlay:
                # Set Overlay
                proflow.overlay_state = context.space_data.overlay.show_overlays
                proflow.overlay_wire_state = context.space_data.overlay.show_wireframes
                context.space_data.overlay.show_overlays = True
                context.space_data.overlay.show_wireframes = True

            shading = context.space_data.shading
            shading.show_xray = False

            self.act_obj = context.active_object


            if proflow.custom_mesh is None:
                proflow.use_custom_mesh = False

            # Check if Custom mesh is instance
            if proflow.use_custom_mesh :
                if proflow.custom_mesh.type == 'EMPTY':
                    proflow.use_plane_instance = True

            # Material
            if self.act_obj:
                if len(self.act_obj.material_slots) and addon_pref.copy_material:
                    self.mat = self.act_obj.active_material
                    proflow.act_obj_material_name = self.mat.name
                else:
                    proflow.act_obj_material_name = ""
                    proflow.replace_asset = False
                    proflow.no_material = True
            else:
                proflow.no_material = True

            bpy.ops.object.duplicate_move()
            bpy.ops.object.convert(target='MESH')
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
            bpy.ops.object.mode_set(mode='EDIT')
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = True
            else:
                context.space_data.overlay.show_retopology = True
            # bpy.context.space_data.overlay.show_occlude_wire = True
            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
            bpy.ops.mesh.select_all(action='DESELECT')

            self.start = False

            context.window_manager.modal_handler_add(self)
            MyHandleClass.remove_handler()
            MyHandleClass.add_handler(proflow_draw_callback_px,(x, y, ps, context.object))

            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}

class PROFLOW_OT_modal_extract(Operator):
    """
    FLOATERS FROM EDGES

    From inside the Modal, select some Edges to make The Flow.

    SPACE - Extract curve ad make Welfing from selected Edges
    CTRL + SPACE - Extract Edge only
    """
    bl_idname = "object.proflow_modal_extract"
    bl_label = "Proflow Extract"

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        if len([obj for obj in context.selected_objects if context.object is not None if obj.type == 'MESH']) == 1:
            return True

    lasso = False

    # MODAL
    def modal(self, context, event):
        proflow = context.window_manager.proflow
        addon_pref = get_addon_preferences()

        if context.area is None:
            context.area.tag_redraw()
            return {'FINISHED'}

        if event.type in {'NUMPAD_1', 'NUMPAD_3', 'NUMPAD_7', 'NUMPAD_5'} or event.ctrl and event.type in {'NUMPAD_1',
                                                                                                           'NUMPAD_3',
                                                                                                           'NUMPAD_7',
                                                                                                           'NUMPAD_5'}:
            return {'PASS_THROUGH'}

        if event.type in {'CTRL', 'SHIFT','ALT', 'LEFTMOUSE', 'R', 'I', 'G', 'ONE', 'TWO', 'THREE', 'C', 'B'}:
            return {'PASS_THROUGH'}


        if event.type == 'MIDDLEMOUSE' or event.alt and event.type == 'MIDDLEMOUSE':
            return {'PASS_THROUGH'}

        if event.type == 'LEFTMOUSE' and event.alt or event.type == 'RIGHTMOUSE' and event.alt:
            return {'PASS_THROUGH'}

        if event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} or event.ctrl and event.type in {'WHEELUPMOUSE',
                                                                                             'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        elif event.ctrl and event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'}:

            context.area.tag_redraw()
            MyHandleClass.remove_handler()
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = False
            else:
                context.space_data.overlay.show_retopology = False
            # bpy.context.space_data.overlay.show_occlude_wire = False
            bpy.context.object.show_in_front = False

            me = context.object.data
            bm = bmesh.from_edit_mesh(me)

            if any([e.select for e in bm.faces]):
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.delete(type='FACE')
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.dissolve_mode()
                bpy.ops.mesh.delete(type='ONLY_FACE')
                bpy.ops.object.mode_set(mode='OBJECT')

                if proflow.is_set_replace_asset:
                    proflow.replace_asset = True
                if proflow.is_set_mat:
                    addon_pref.copy_material = True
                if addon_pref.edit_overlay:
                    # restore overlay
                    context.space_data.overlay.show_overlays = True
                    context.space_data.overlay.show_wireframes = True
                self.act_obj.hide_set(False)
                return {'FINISHED'}

            elif any([e.select for e in bm.edges]):
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.delete(type='VERT')
                bpy.ops.object.mode_set(mode='OBJECT')
                if proflow.is_set_replace_asset:
                    proflow.replace_asset = True
                if proflow.is_set_mat:
                    addon_pref.copy_material = True
                if addon_pref.edit_overlay:
                    # restore overlay
                    context.space_data.overlay.show_overlays = True
                    context.space_data.overlay.show_wireframes = True
                    self.act_obj.hide_set(False)
                return {'FINISHED'}
            else:
                self.report({'WARNING'}, "Select some edges")

        # Stop and finalize
        elif event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'}:
            context.area.tag_redraw()
            MyHandleClass.remove_handler()

            self.act_obj.hide_set(False)

            me = context.object.data
            bm = bmesh.from_edit_mesh(me)

            if any([e.select for e in bm.edges]):
                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
                bpy.ops.mesh.select_all(action='INVERT')
                bpy.ops.mesh.delete(type='VERT')
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.transform_apply(location=False, rotation=True, scale=False) # bug en poly à cause du apply

                self.mesh_to_cut = context.active_object
                if bpy.app.version < (3, 6, 0):
                    context.space_data.overlay.show_occlude_wire = False
                else:
                    context.space_data.overlay.show_retopology = False
                # bpy.context.space_data.overlay.show_occlude_wire = False
                bpy.ops.object.convert(target='CURVE')

                bpy.ops.object.proflow_finalize()

                if proflow.is_set_replace_asset:
                    proflow.replace_asset = True
                if proflow.is_set_mat:
                    addon_pref.copy_material = True

                proflow.from_modals = True
                return {'FINISHED'}
            else:
                self.report({'WARNING'}, "Select some edges")

        # CANCEL AND EXIT
        elif event.type in {'ESC', 'RIGHTMOUSE'} and event.value == 'PRESS':
        # elif event.type == 'ESC' and event.value == 'PRESS':
            MyHandleClass.remove_handler()
            context.area.tag_redraw()
            if bpy.app.version < (3, 6, 0):
                context.space_data.overlay.show_occlude_wire = False
            else:
                context.space_data.overlay.show_retopology = False
            # bpy.context.space_data.overlay.show_occlude_wire = False
            bpy.context.object.show_in_front = False
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.delete(use_global=False)

            self.act_obj.hide_set(False)
            context.view_layer.objects.active = self.act_obj
            self.act_obj.select_set(state=True)
            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True

            if addon_pref.edit_overlay:
                # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):

        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "INFO: Do not work in Local Mode")
        #             return {'FINISHED'}

        if context.area.type == 'VIEW_3D' and context.object:
            proflow = context.window_manager.proflow
            addon_pref = get_addon_preferences()

            if addon_pref.edit_overlay:
                # Set Overlay
                proflow.overlay_state = context.space_data.overlay.show_overlays
                proflow.overlay_wire_state = context.space_data.overlay.show_wireframes
                context.space_data.overlay.show_overlays = True
                context.space_data.overlay.show_wireframes = True

            shading = context.space_data.shading
            shading.show_xray = False

            self.act_obj = context.active_object

            if proflow.custom_mesh is None:
                proflow.use_custom_mesh = False

            # Check if Custom mesh is instance
            if proflow.use_custom_mesh:
                if proflow.custom_mesh.type == 'EMPTY':
                    proflow.use_plane_instance = True

            proflow.active_mesh = self.act_obj


            # Material
            if self.act_obj:
                if len(self.act_obj.material_slots) and addon_pref.copy_material:
                    self.mat = self.act_obj.active_material
                    proflow.act_obj_material_name = self.mat.name
                else:
                    proflow.act_obj_material_name = ""
                    proflow.replace_asset = False
                    proflow.no_material = True
            else:
                proflow.no_material = True

            if len([o for o in context.selected_objects]) == 1:

                bpy.ops.object.duplicate_move()
                self.new_object = context.active_object

                if not event.shift:
                    self.act_obj.hide_set(True)

                self.width_value = 0
                unit = bpy.context.scene.unit_settings.scale_length
                if np.isclose(1, unit):
                    self.width_value = 0.002
                elif np.isclose(0.01, unit):
                    self.width_value = 0.02
                elif np.isclose(0.001, unit):
                    self.width_value = 0.2

                if self.new_object.modifiers:
                    for mod in self.new_object.modifiers:
                        if mod.type == 'BEVEL':
                            if mod.limit_method == 'ANGLE':
                                if mod.width <= self.width_value:
                                    bpy.ops.object.modifier_remove(modifier=mod.name)


                bpy.ops.object.convert(target='MESH')
                bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
                context.object.show_in_front = True
                if bpy.app.version < (4, 1, 0):
                    context.object.data.use_auto_smooth = False
                bpy.ops.object.shade_flat()
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.customdata_custom_splitnormals_clear()

                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='FACE')
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.transform.edge_bevelweight(value=-1)
                bpy.ops.transform.edge_crease(value=-1)
                bpy.ops.mesh.mark_sharp(clear=True)
                bpy.ops.mesh.mark_seam(clear=True)
                bpy.ops.mesh.select_all(action='DESELECT')

                if event.shift:
                    return {'FINISHED'}

                bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')

            context.window_manager.modal_handler_add(self)
            MyHandleClass.remove_handler()
            MyHandleClass.add_handler(proflow_draw_callback_px, (x, y, ps1, context.object))
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}


tmp_verts = []
tmp_verts_storage = []

class PROFLOW_OT_scale(bpy.types.Operator):
    """Scale mesh with mouse"""
    bl_idname = "object.proflow_scale"
    bl_label = "Simple Modal Operator"
    bl_options = {'GRAB_CURSOR','BLOCKING'}

    first_mouse_x: IntProperty()
    first_value: FloatProperty()
    action_scale = False

    obj_matrix_save = None
    start_end_cap = False
    random = False

    def modal(self, context, event):
        addon_pref = get_addon_preferences()

        if event.type == 'MOUSEMOVE' :
            speed = 0.01 if event.shift else 0.5

            delta = 1. - 0.01 * (self.first_mouse_x - event.mouse_x) * speed * addon_pref.modal_speed


            if  self.is_instance:
                delta = (event.mouse_x - self.first_mouse_x) * 0.005 * speed
                MS = self.MW @ Matrix.Scale(1 + delta, 4)  # 1 + delta parce que c'est un mutliplicateur.
                # Vue que le delta commence toujours à 0 au lancement du modal,
                # le scale de l'objet sera multiplié par 0.
                # Pour éviter ça, on ajoute 1 ce qui permet de ne pas changer
                # le scale au lancement du modal.
                context.object.matrix_world = MS

            elif self.is_special_array:
                for mesh in self.meshes:
                    obj = mesh["obj"]
                    matrix = mesh["matrix"]

                    delta = 0.05 * (event.mouse_x - self.first_mouse_x) * speed / 100
                    obj.matrix_world = matrix @ Matrix.Scale(1 + delta, 4)
                # delta = (event.mouse_x - self.first_mouse_x) * 0.005 * speed
                # MS_act_obj = self.act_obj @ Matrix.Scale(1 + delta, 4)
                # self.act_obj.matrix_world = MS_act_obj
                #
                # if self.array_start_cap:
                #     MS_start_cap = self.array_start_cap @ Matrix.Scale(1 + delta, 4)
                #     self.array_start_cap.matrix_world = MS_start_cap
                #
                # if self.array_end_cap:
                #     MS_end_cap = self.array_end_cap @ Matrix.Scale(1 + delta, 4)
                #     self.array_end_cap.matrix_world = MS_end_cap
                #
                # for mod in self.act_obj.modifiers:
                #     if mod.type == 'BEVEL':
                #         if self.bevels.get(mod.name):
                #             if event.ctrl:
                #                 mod.width = delta* self.bevels[mod.name]# on recup la valeur qu'on a stoqué
                #             else:
                #                 mod.width = (delta - delta % 0.1) * self.bevels[mod.name]


            else:
                for i, v in enumerate(context.object.data.vertices):
                    if event.ctrl:
                        v.co.x = delta* tmp_verts[i][0]
                        v.co.y = delta* tmp_verts[i][1]
                        v.co.z = delta* tmp_verts[i][2]
                    else:
                        v.co.x = (delta - delta % 0.1) * tmp_verts[i][0]
                        v.co.y = (delta - delta % 0.1) * tmp_verts[i][1]
                        v.co.z = (delta - delta % 0.1) * tmp_verts[i][2]

                for mod in self.act_obj.modifiers:
                    if mod.type == 'BEVEL':
                        if self.bevels.get(mod.name):
                            if event.ctrl:
                                mod.width = delta* self.bevels[mod.name]# on recup la valeur qu'on a stoqué
                            else:
                                mod.width = (delta - delta % 0.1) * self.bevels[mod.name]


        elif event.type in {'LEFTMOUSE', 'SPACE'} :

            if self.is_special_array:
                bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

            self.is_special_array = False
            self.is_instance = False
            bpy.ops.object.select_all(action='DESELECT')

            context.view_layer.objects.active = self.act_obj
            self.act_obj.select_set(state=True)

            if self.subsurf_show_viewport:
                self.act_mod.show_viewport = True

            # Hide start et end cap
            if context.object.type == 'MESH':
                for mod in context.object.modifiers:
                    if mod.type == 'ARRAY':
                        if mod.start_cap is not None:
                            self.array_start_cap = mod.start_cap
                            self.array_start_cap.hide_set(True)
                        if mod.end_cap is not None:
                            self.array_end_cap = mod.end_cap
                            self.array_end_cap.hide_set(True)

            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            self.is_instance = False

            context.view_layer.objects.active = self.act_obj
            self.act_obj.select_set(state=True)

            if self.subsurf_show_viewport:
                self.act_mod.show_viewport = True
            context.object.location.x = self.first_value
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        self.act_obj = context.active_object

        self.subsurf = False
        self.subsurf_show_viewport = False
        self.is_instance = False
        self.is_special_array = False
        self.is_bound = False

        self.sel = []

        if self.act_obj.name.startswith("proflow_plane_instance"):
            self.is_instance = True

        if self.act_obj.modifiers:
            for mod in self.act_obj.modifiers:
                if mod.type == 'SUBSURF':
                    self.subsurf = True
                    self.act_mod = mod

                if mod.type == 'ARRAY':
                    if mod.start_cap is not None:
                        self.is_special_array = True
                        self.array_start_cap = mod.start_cap
                        self.sel.append(self.array_start_cap)
                        self.array_start_cap.hide_set(False)

                    if mod.end_cap is not None:
                        self.is_special_array = True
                        self.array_end_cap = mod.end_cap
                        self.sel.append(self.array_end_cap)
                        self.array_end_cap.hide_set(False)



        # print(self.sel)


        if not self.is_instance and not self.is_special_array:
            self.first_mouse_x = event.mouse_x
            self.first_value = context.object.location.x
            while len(tmp_verts)>0:
                del tmp_verts[0]
            for v in context.object.data.vertices:
                tmp_verts.append([v.co.x, v.co.y, v.co.z])
                tmp_verts_storage.append([v.co.x, v.co.y, v.co.z])


            if self.subsurf:
                if self.act_mod.show_viewport:
                    self.subsurf_show_viewport = True
                    self.act_mod.show_viewport = False

            self.bevels = {}
            for mod in self.act_obj.modifiers:
                if mod.type == 'BEVEL':
                    self.bevels[mod.name] = mod.width

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        elif self.is_special_array:

            if self.subsurf:
                if self.act_mod.show_viewport:
                    self.subsurf_show_viewport = True
                    self.act_mod.show_viewport = False

            # self.bevels = {}
            # for mod in self.act_obj.modifiers:
            #     if mod.type == 'BEVEL':
            #         self.bevels[mod.name] = mod.width


            # bpy.ops.object.select_all(action='DESELECT')

            # for obj in self.sel:
            #     obj.select_set(state=True)
            #     return {'FINISHED'}
            self.sel.append(self.act_obj)

            self.meshes = []
            for obj in self.sel:
                obj.select_set(state=True)
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy()
                })


            # self.MW_act_obj = context.object.matrix_world.copy()  # on copie la matrice pour l'avoir en reference

            self.first_mouse_x = event.mouse_x
            # self.first_value = sum(self.act_obj.scale[:]) / 3


            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}


        elif self.is_instance:
            bpy.ops.object.select_grouped(type='CHILDREN')

            for obj in context.selected_objects:
                context.view_layer.objects.active = obj
                self.act_obj = context.active_object
                self.instance = context.active_object

            self.MW = context.object.matrix_world.copy()  # on copie la matrice pour l'avoir en reference
            self.first_mouse_x = event.mouse_x
            self.first_value = sum(context.object.scale[:]) / 3

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        else:
            self.report({'WARNING'}, "No active object, could not finish")
            return {'CANCELLED'}


def local_rotate(obj, axis, angle):
    obj.matrix_world @= Matrix.Rotation(math.radians(angle), 4, axis)

INCREMENT = 10
class PROFLOW_OT_modal_rotate_scale(Operator):
    """EDIT ASSET WITH MODAL"""
    bl_idname = "object.proflow_modal_rotate_scale"
    bl_label = "Rotate and Scale"
    bl_options = {'GRAB_CURSOR','BLOCKING'}

    first_mouse_x: IntProperty()
    first_value: FloatProperty()

    axis_index = 0
    act_obj_matrix_save = None
    child_instance_matrix_save = None
    profile_matrix_save = None
    rotate = False
    rotate_instance = False
    value_rotation = 0
    edit_displace_1 = False
    edit_displace_2 = False
    edit_bend = False
    edit_array_space = False
    array_direction = 2
    curve_offset = False

    # Curve
    curve_scale = False
    curve_resolution = False
    profile_resolution = False
    is_profile_curve = False
    profile_rotation = False
    profile_bevel_end = False
    profile_bevel_start = False


    from_panel : BoolProperty(default=False)

    def toggle_bezier_poly(self, context):

        obj = context.active_object
        context.view_layer.objects.active = obj
        obj.select_set(state=True)
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')

        if obj.data.splines[0].type == 'BEZIER':
            bpy.ops.curve.spline_type_set(type='POLY')
        else:
            bpy.ops.curve.spline_type_set(type='BEZIER')
            bpy.ops.curve.handle_type_set(type='VECTOR')

        bpy.ops.object.mode_set(mode='OBJECT')

    def modal_tilt(self):
        bpy.ops.transform.tilt('INVOKE_DEFAULT', release_confirm=True)

    # MODAL
    def modal(self, context, event):
        proflow = context.window_manager.proflow
        addon_pref = get_addon_preferences()

        context.area.tag_redraw()

        if event.type in {'NUMPAD_1', 'NUMPAD_3', 'NUMPAD_7', 'NUMPAD_5'} or event.ctrl and event.type in {'NUMPAD_1',
                                                                                                           'NUMPAD_3',
                                                                                                           'NUMPAD_7',
                                                                                                           'NUMPAD_5'}:
            return {'PASS_THROUGH'}

        if event.type in {'CTRL', 'SHIFT', 'ALT'}:
            return {'PASS_THROUGH'}

        if event.type == 'MIDDLEMOUSE' or event.alt and event.type == 'MIDDLEMOUSE':
            return {'PASS_THROUGH'}

        if event.type == 'LEFTMOUSE' and event.alt or event.type == 'RIGHTMOUSE' and event.alt:
            return {'PASS_THROUGH'}

        if event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} or event.ctrl and event.type in {'WHEELUPMOUSE',
                                                                                             'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}


        if event.type == 'MOUSEMOVE':

            speed = 0.1 if event.shift else 0.5

            # if event.ctrl and event.shift
            #
            # if event.shift:
            #     speed = 0.1
            # else:
            #     speed = 0.5



            delta = (self.first_mouse_x - event.mouse_x) * speed
            # speed = 0.1 if event.shift else 0.5
            # delta = 1. - 0.01 * (self.first_mouse_x - event.mouse_x) * speed

            if self.act_obj.type == 'MESH':
                if self.rotate:
                    self.act_obj.matrix_world = self.act_obj_matrix_save

                    if self.act_mod_array.relative_offset_displace[0] > 0:
                        self.angle = 'X'
                    elif self.act_mod_array.relative_offset_displace[1] > 0:
                        self.angle = 'Y'
                    elif self.act_mod_array.relative_offset_displace[2] > 0:
                        self.angle = 'Z'

                    if event.ctrl:
                        self.act_obj.matrix_world @= Matrix.Rotation(radians(delta * addon_pref.modal_speed), 4, self.angle)
                    else:
                        self.act_obj.matrix_world @=  Matrix.Rotation(radians(delta - delta % 45), 4, self.angle)

                # ROTATE INSTANCE
                elif self.rotate_instance:
                    self.child_instance.matrix_world = self.child_instance_matrix_save
                    if event.ctrl:
                        self.child_instance.matrix_world @= Matrix.Rotation(radians(delta * addon_pref.modal_speed), 4, 'X')
                    else:
                        self.child_instance.matrix_world @=  Matrix.Rotation(radians(delta - delta % 45), 4, 'X')

                elif self.edit_displace_1:
                    self.act_mod_disp_1.strength = self.value_disp_1 + delta /250  * addon_pref.modal_speed

                elif self.edit_displace_2:
                    self.act_mod_disp_2.strength = self.value_disp_2 + delta /250  * addon_pref.modal_speed

                elif self.edit_bend:
                    self.act_mod_bend.angle = self.value_bend + radians(delta) * 0.7  * addon_pref.modal_speed

                elif self.edit_array_space:
                    if self.array_direction == 0:
                        self.act_mod_array.relative_offset_displace[0] = self.value_array_space - delta /100  * addon_pref.modal_speed
                    elif self.array_direction == 1:
                        self.act_mod_array.relative_offset_displace[1] = self.value_array_space - delta /100  * addon_pref.modal_speed
                    elif self.array_direction == 2:
                        self.act_mod_array.relative_offset_displace[2] = self.value_array_space - delta /100  * addon_pref.modal_speed

            elif self.act_obj.type == 'CURVE':
                act_data = self.act_obj.data

                # SCALE PROFILE
                if act_data.bevel_object is not None and self.curve_scale:
                    MS = self.MW @ Matrix.Scale(1 - delta/100  * addon_pref.modal_speed, 4)
                    self.profile.matrix_world = MS

                elif act_data.bevel_object is not None and self.profile_resolution:
                    act_data.bevel_object.data.resolution_u = self.first_value - delta / 100  * addon_pref.modal_speed

                if self.curve_resolution:
                    act_data.resolution_u = self.first_value - delta /100  * addon_pref.modal_speed

                if self.curve_offset:
                    act_data.offset = self.first_value + delta / 100  * addon_pref.modal_speed

                # ROTATE PROFILE
                # if self.rotate:
                #     # self.modal_tilt()
                #     bpy.ops.transform.tilt('INVOKE_DEFAULT')

                    # Reinit
                    # self.reinit()
                    # Delta
                    # delta = 0.01 * (self.first_mouse_x - event.mouse_x)

                    # Incremented
                    # delta -= delta%(INCREMENT*(pi / 180.))

                    # delta -= radians(delta - delta % 45)
                    # delta = radians(delta - delta % 10)


                    # Rotate
                    # bpy.ops.transform.rotate(value=delta, orient_axis='Z', orient_type='GLOBAL',
                    #                          orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
                    #                          orient_matrix_type='GLOBAL', constraint_axis=(False, False, True))
                    # bpy.ops.object.mode_set(mode='EDIT')

                    # bpy.ops.transform.tilt(value=radians(delta - delta % 45) * 0.01)
                    # bpy.ops.transform.tilt('INVOKE_DEFAULT', True)

                    # bpy.ops.object.mode_set(mode='OBJECT')


                if self.profile_bevel_start:
                    act_data.bevel_factor_start = self.first_value - delta / 250  * addon_pref.modal_speed

                if self.profile_bevel_end:
                    act_data.bevel_factor_end = self.first_value - delta / 250  * addon_pref.modal_speed

        if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
            if context.object.type == 'CURVE' and self.rotate and self.profile_rotation:

                bpy.ops.object.mode_set(mode='OBJECT')
                # bpy.ops.object.select_all(action='DESELECT')
                # self.act_obj.select_set(state=True)
                # context.view_layer.objects.active = self.act_obj
                # bpy.ops.object.mode_set(mode='OBJECT')
                # context.scene.tool_settings.transform_pivot_point = self.pivot


            self.rotate = False
            self.rotate_instance = False
            self.edit_displace_1 = False
            self.edit_displace_2 = False
            self.edit_bend = False
            self.edit_array_space = False
            self.curve_offset = False
            self.pause = True
            self.profile_bevel_start = False
            self.profile_bevel_end = False


            # CURVE
            self.curve_scale = False
            self.profile_resolution = False
            self.curve_resolution = False
            self.profile_rotation = False

        if event.type == 'S' and event.value == 'PRESS':
            self.first_mouse_x = event.mouse_x
            if self.act_obj.type == 'MESH':
                bpy.ops.object.proflow_scale("INVOKE_DEFAULT")
                self.pause = True

            else:
                self.curve_scale = True
                self.profile = self.act_obj.data.bevel_object
                self.MW = self.profile.matrix_world.copy()
                self.first_value = sum(self.profile.scale[:]) / 3

        # ROTATE INSTANCE
        elif event.ctrl and event.type == 'R' and event.value == 'PRESS' and self.has_instance:
            self.first_mouse_x = event.mouse_x
            self.rotate_instance = True

        elif event.type == 'R' and event.value == 'PRESS':
            self.first_mouse_x = event.mouse_x
            self.rotate = True

            # if self.act_obj.type == 'CURVE':
            #     self.pivot = context.scene.tool_settings.transform_pivot_point
            #
            #     self.profile_rotation = True
            #     # self.profile = self.act_obj.data.bevel_object
            #
            #     # bpy.ops.object.select_all(action='DESELECT')
            #     # context.view_layer.objects.active = self.profile
            #     # self.profile.select_set(state=True)
            #     # self.pts = [list(pt.co) for pt in self.profile.data.splines[0].points]
            #     # bpy.ops.view3d.snap_cursor_to_selected()
            #     # bpy.context.scene.tool_settings.transform_pivot_point = 'CURSOR'
            #
            #     bpy.ops.object.mode_set(mode='EDIT')
            #     bpy.ops.curve.select_all(action='SELECT')


        if event.type == 'V' and context.object.type == 'CURVE':
            self.first_mouse_x = event.mouse_x
            self.profile_bevel_start = True
            self.first_value = self.act_obj.data.bevel_factor_start

        if event.type == 'B' and context.object.type == 'CURVE':
            self.first_mouse_x = event.mouse_x
            self.profile_bevel_end = True
            self.first_value = self.act_obj.data.bevel_factor_end

        if event.type == 'E' and context.object.type == 'CURVE':
            self.first_mouse_x = event.mouse_x
            self.curve_offset = True
            self.first_value = self.act_obj.data.offset

        # DISPLACE X
        elif event.type == 'D' and event.value == 'PRESS':
            self.first_mouse_x = event.mouse_x
            if self.act_obj.type == 'MESH':
                if self.displace_1:
                    self.edit_displace_1 = True
                    self.value_disp_1 = self.act_mod_disp_1.strength

            elif self.act_obj.type == 'CURVE':
                self.curve_resolution = True
                self.first_value = self.act_obj.data.resolution_u

        # DISPLACE Y
        elif event.type == 'G' and event.value == 'PRESS':
            self.first_mouse_x = event.mouse_x

            if self.act_obj.type == 'MESH':
                if self.displace_2:
                    self.edit_displace_2 = True
                    self.value_disp_2 = self.act_mod_disp_2.strength

            elif self.act_obj.type == 'CURVE':
                self.profile_resolution = True
                self.profile = self.act_obj.data.bevel_object
                self.first_value = self.profile.data.resolution_u

        # BEND
        elif event.type == 'B' and event.value == 'PRESS' and self.act_obj.type == 'MESH':
            self.first_mouse_x = event.mouse_x
            if self.bend:
                self.edit_bend = True
                self.act_mod_bend.show_viewport = True
                self.value_bend = self.act_mod_bend.angle

        # ARRAY SPACE
        elif event.type == 'Q' and event.value == 'PRESS' and not self.is_profile_curve:
            self.first_mouse_x = event.mouse_x
            self.edit_array_space = True

            if self.act_mod_array.relative_offset_displace[0] > 0 :
                self.value_array_space = self.act_mod_array.relative_offset_displace[0]
                self.array_direction = 0
            elif self.act_mod_array.relative_offset_displace[1] > 0 :
                self.value_array_space = self.act_mod_array.relative_offset_displace[1]
                self.array_direction = 1
            elif self.act_mod_array.relative_offset_displace[2] > 0 :
                self.value_array_space = self.act_mod_array.relative_offset_displace[2]
                self.array_direction = 2

        if event.type == 'H' and event.value == 'PRESS' and self.is_profile_curve:
            context.object.data.use_fill_caps = not context.object.data.use_fill_caps

        # CYCLIC
        if event.type == 'C' and event.value == 'PRESS' :

            if self.act_obj.data.splines[0].use_cyclic_u == False:
                self.act_obj.data.splines[0].use_cyclic_u = True
                self.report({'INFO'}, "Cyclic ON")
            else:
                self.act_obj.data.splines[0].use_cyclic_u = False
                self.report({'INFO'}, "Cyclic OFF")

        if event.type == 'X' and event.value == 'PRESS':
            self.rotate = False
            proflow.directions = 'x'

        if event.type == 'Y' and event.value == 'PRESS':
            self.rotate = False
            proflow.directions = 'y'

        if event.type == 'Z' and event.value == 'PRESS':
            self.rotate = False
            proflow.directions = 'z'

        if event.ctrl and event.type == 'A' and event.value == 'PRESS':
            bpy.ops.object.proflow_convert_to_mesh()
            MyHandleClass.remove_handler()

            context.area.tag_redraw()
            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True


            proflow.on_curve = False
            proflow.from_modals = False
            self.is_profile_curve = False
            if addon_pref.edit_overlay:
                # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state

            return {'FINISHED'}

        if event.type == 'P' and event.value == 'PRESS':
            if context.object.type == 'MESH':
                bpy.ops.object.select_all(action='DESELECT')
                context.view_layer.objects.active = self.curve
                self.curve.select_set(state=True)

                self.toggle_bezier_poly(context)

                bpy.ops.object.select_all(action='DESELECT')
                context.view_layer.objects.active = self.act_obj
                self.act_obj.select_set(state=True)

            elif context.object.type == 'CURVE':
                self.toggle_bezier_poly(context)

        # FLIP CURVE DIRECTION
        if event.type == 'F' and event.value == 'PRESS':
            if self.act_obj.type == 'MESH':
                bpy.ops.object.select_all(action='DESELECT')
                context.view_layer.objects.active = self.curve
                self.curve.select_set(state=True)
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.curve.select_all(action='SELECT')
                bpy.ops.curve.switch_direction()
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.select_all(action='DESELECT')
                context.view_layer.objects.active = self.act_obj
                self.act_obj.select_set(state=True)
            elif self.act_obj.type == 'CURVE':
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.curve.switch_direction()
                bpy.ops.object.mode_set(mode='OBJECT')

        # Stop and finalize
        if event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'} and self.pause:
            MyHandleClass.remove_handler()
            context.area.tag_redraw()

            context.view_layer.objects.active = self.act_obj
            self.act_obj.select_set(state=True)

            if self.act_obj.type in {'MESH', 'FONT'} and not self.has_instance:

                # Hide start et end cap
                for mod in context.object.modifiers:
                    if mod.type == 'ARRAY':
                        if mod.start_cap is not None:
                            self.array_start_cap = mod.start_cap
                            self.array_start_cap.hide_set(True)
                        if mod.end_cap is not None:
                            self.array_end_cap = mod.end_cap
                            self.array_end_cap.hide_set(True)

                if self.subsurf :
                    self.act_mod_subsurf.show_viewport = True

                if context.object.type == 'MESH' and self.weld:
                    if self.weld_viewport is not False:
                        self.act_mod_weld.show_viewport = True
            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True

            proflow.on_curve = False
            proflow.from_modals = False
            self.is_profile_curve = False

            if addon_pref.edit_overlay:
            # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state
            return {'FINISHED'}

        # CANCEL AND EXIT
        if event.type in {'RIGHTMOUSE', 'ESC'}:
        # if event.type == 'ESC' and event.value == 'PRESS':
            MyHandleClass.remove_handler()
            context.area.tag_redraw()

            # if self.profile_rotation:
            #     self.reinit()

            if context.object.type == 'CURVE':
                if context.object.mode == 'EDIT':
                    bpy.ops.object.mode_set(mode='OBJECT')
                    bpy.ops.object.select_all(action='DESELECT')
                    context.view_layer.objects.active = self.act_obj
                    self.act_obj.select_set(state=True)

            if self.act_obj.type == 'MESH':

                # UNHide start et end cap
                for mod in context.object.modifiers:
                    if mod.type == 'ARRAY':
                        if mod.start_cap is not None:
                            self.array_start_cap = mod.start_cap
                            self.array_start_cap.hide_set(True)
                        if mod.end_cap is not None:
                            self.array_end_cap = mod.end_cap
                            self.array_end_cap.hide_set(True)

                # Restore  Values
                if self.displace_1:
                    self.act_mod_disp_1.strength = self.dis_1_strength
                if self.displace_2:
                    self.act_mod_disp_2.strength = self.dis_2_strength
                if self.bend:
                    self.act_mod_bend.angle = self.bend_strength
                if self.array:
                    self.act_mod_array.relative_offset_displace[0] = self.arra_space_x
                    self.act_mod_array.relative_offset_displace[1] = self.arra_space_y
                    self.act_mod_array.relative_offset_displace[2] = self.arra_space_z

                self.act_obj.matrix_world = self.act_obj_matrix_save

                for i, v in enumerate(context.object.data.vertices):
                    v.co.x = tmp_verts_storage[i][0]
                    v.co.y = tmp_verts_storage[i][1]
                    v.co.z = tmp_verts_storage[i][2]

            if self.act_obj.type in {'MESH', 'FONT'} and not self.has_instance:
                if self.subsurf:
                    self.act_mod_subsurf.show_viewport = True

                if context.object.type == 'MESH' and self.weld:
                    if self.weld_viewport is not False:
                        self.act_mod_weld.show_viewport = True

            if proflow.is_set_replace_asset:
                proflow.replace_asset = True
            if proflow.is_set_mat:
                addon_pref.copy_material = True

            proflow.on_curve = False
            proflow.from_modals = False
            self.is_profile_curve = False
            if addon_pref.edit_overlay:
                # restore overlay
                context.space_data.overlay.show_overlays = proflow.overlay_state
                context.space_data.overlay.show_wireframes = proflow.overlay_wire_state
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    # def reinit(self):
    #     # Put all vertices in the original state
    #     for i, pt in enumerate(self.pts):
    #         self.profile.data.splines[0].points[i].co.x = pt[0]
    #         self.profile.data.splines[0].points[i].co.y = pt[1]
    #         self.profile.data.splines[0].points[i].co.z = pt[2]

    def invoke(self, context, event):

        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "INFO: Do not work in Local Mode")
        #             return {'FINISHED'}


        if context.area.type == 'VIEW_3D' and context.object:
            self.first_mouse_x = event.mouse_x
            self.first_mouse_y = event.mouse_y
            proflow = context.window_manager.proflow
            addon_pref = get_addon_preferences()

            # Set Overlay
            if addon_pref.edit_overlay:
                if self.from_panel or proflow.from_modals:
                    proflow.overlay_state = context.space_data.overlay.show_overlays
                    proflow.overlay_wire_state = context.space_data.overlay.show_wireframes
                else:
                    context.space_data.overlay.show_overlays = proflow.overlay_state
                    context.space_data.overlay.show_wireframes = proflow.overlay_wire_state

                if proflow.keep_wireframe:
                    context.space_data.overlay.show_overlays = True
                    context.space_data.overlay.show_wireframes = True


            # Set Mesh
            self.act_obj = context.active_object

            # Check Subsurf
            self.subsurf = False
            for mod in context.object.modifiers:
                if mod.type == 'SUBSURF':
                    self.subsurf = True
                    self.act_mod_subsurf = mod
                    if mod.show_viewport:
                        self.subsurf_show_viewport = True

            # Set Curve
            if self.act_obj.type == 'MESH':
                if self.act_obj.modifiers:
                    for mod in self.act_obj.modifiers:
                        if mod.type == 'CURVE' and mod.name == 'PFLOW - Curve':
                            self.curve = mod.object
                            self.act_spline = self.curve.data.splines.active

                        #UNHide start et end cap
                        if mod.type == 'ARRAY':
                            if mod.start_cap is not None:
                                self.array_start_cap = mod.start_cap
                                self.array_start_cap.hide_set(False)
                            if mod.end_cap is not None:
                                self.array_end_cap = mod.end_cap
                                self.array_end_cap.hide_set(False)

                # Check subsurf

                self.displace_1 = False
                self.displace_2 = False
                self.bend = False
                self.subsurf_show_viewport = False
                self.array = False
                self.weld = False
                for mod in context.object.modifiers:
                    if mod.type == 'DISPLACE' and mod.name == 'PFLOW - Displace - 1':
                        self.displace_1 = True
                        self.act_mod_disp_1 = mod
                    if mod.type == 'DISPLACE' and mod.name == 'PFLOW - Displace - 2':
                        self.displace_2 = True
                        self.act_mod_disp_2 = mod
                    if mod.type == 'SIMPLE_DEFORM' and mod.name == 'PFLOW - Bend':
                        self.bend = True
                        self.act_mod_bend = mod
                    if mod.type == 'ARRAY' and mod.name == 'PFLOW - Array':
                        self.array = True
                        self.act_mod_array = mod
                    if mod.type == 'WELD' and mod.name == 'PFLOW - Weld':
                        self.weld = True
                        self.act_mod_weld = mod

                self.has_instance = False
                if self.act_obj.name.startswith("proflow_plane_instance"):
                    if self.act_obj.children:
                        self.has_instance = True
                        for child in self.act_obj.children:
                            self.child_instance = child

                # store current Values
                if self.has_instance:
                    self.child_instance_matrix_save = self.child_instance.matrix_world.copy()

                self.act_obj_matrix_save = self.act_obj.matrix_world.copy()

                if self.displace_1:
                    self.dis_1_strength = self.act_mod_disp_1.strength
                if self.displace_2:
                    self.dis_2_strength = self.act_mod_disp_2.strength
                if self.bend:
                    self.bend_strength = self.act_mod_bend.angle
                if self.array:
                    self.arra_space_x = self.act_mod_array.relative_offset_displace[0]
                    self.arra_space_y = self.act_mod_array.relative_offset_displace[1]
                    self.arra_space_z = self.act_mod_array.relative_offset_displace[2]


                while len(tmp_verts_storage)>0:
                    del tmp_verts_storage[0]
                for v in context.object.data.vertices:
                    tmp_verts_storage.append([v.co.x, v.co.y, v.co.z])

            elif self.act_obj.type == 'CURVE' and context.object.data.bevel_object is not None:

                self.is_profile_curve = True
                self.profile = context.object.data.bevel_object
                self.first_value = context.object.scale.x

            self.subsurf_viewport = False
            if self.subsurf:
                if self.act_mod_subsurf.show_viewport:
                    self.subsurf_viewport = True
                self.act_mod_subsurf.show_viewport = False


            self.weld_viewport = False
            if context.object.type == 'MESH' and self.weld:
                if self.act_mod_weld.show_viewport:
                    self.weld_viewport = True
                self.act_mod_weld.show_viewport = False


            self.pause = False
            MyHandleClass.remove_handler()
            if self.act_obj.type == 'MESH':
                MyHandleClass.add_handler(proflow_draw_callback_px, (x, y1, ps2, context.object))
            elif self.act_obj.type == 'CURVE':
                MyHandleClass.add_handler(proflow_draw_callback_px, (x, y1, ps3, context.object))
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}


class PROFLOW_OT_loc_rot_scale_random(bpy.types.Operator):
    bl_idname = "object.proflow_loc_rot_scale_random"
    bl_label = "ProFlow Location Rotation Scale Random"

    first_mouse_x: IntProperty()
    first_mouse_y: IntProperty()
    rotate = False
    scale = False
    locate = False
    locate_z = False

    def modal(self, context, event):
        # if context.area is None:
        context.area.tag_redraw()
        # return {'FINISHED'}

        if event.type in {'NUMPAD_1', 'NUMPAD_3', 'NUMPAD_7', 'NUMPAD_5'} or event.ctrl and event.type in {'NUMPAD_1',
                                                                                                           'NUMPAD_3',
                                                                                                           'NUMPAD_7',
                                                                                                           'NUMPAD_5'}:
            return {'PASS_THROUGH'}

        if event.type in {'CTRL', 'SHIFT', 'ALT'}:
            return {'PASS_THROUGH'}

        if event.type == 'MIDDLEMOUSE' or event.alt and event.type == 'MIDDLEMOUSE':
            return {'PASS_THROUGH'}

        if event.type == 'LEFTMOUSE' and event.alt or event.type == 'RIGHTMOUSE' and event.alt:
            return {'PASS_THROUGH'}

        if event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'} or event.ctrl and event.type in {'WHEELUPMOUSE',
                                                                                             'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        if event.type == 'MOUSEMOVE':
            for mesh in self.meshes:
                obj = mesh["obj"]
                matrix = mesh["matrix"]
                speed = mesh["speed"]

                if self.rotate:
                    delta = 0.05 * (event.mouse_x - self.first_mouse_x) * speed / 10
                    obj.matrix_world = matrix @ Matrix.Rotation(delta, 4, 'Z')

                if self.scale:
                    delta = 0.05 * (event.mouse_x - self.first_mouse_x) * speed / 100
                    obj.matrix_world = matrix @ Matrix.Scale(1 + delta, 4)

                if self.locate:
                    delta_x = 0.05 * (event.mouse_x - self.first_mouse_x) * speed / 250
                    delta_y = 0.05 * (event.mouse_y - self.first_mouse_y) * speed / 250
                    obj.matrix_world = matrix @ Matrix.Translation((delta_x, delta_y, 0))

                if self.locate_z:
                    delta_z = 0.05 * (event.mouse_x - self.first_mouse_x) * speed / 250
                    obj.matrix_world = matrix @ Matrix.Translation((0, 0, delta_z))

        elif event.type == 'R' and event.value == 'PRESS':
            for obj in bpy.context.selected_objects:
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy(),
                    "speed": -1. + 2 * random.random()
                })

            self.rotate = True

        elif event.type == 'S' and event.value == 'PRESS':
            for obj in bpy.context.selected_objects:
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy(),
                    "speed": -1. + 2 * random.random()
                })
            self.scale = True

        elif event.type == 'G' and event.value == 'PRESS':
            self.locate = True
            for obj in bpy.context.selected_objects:
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy(),
                    "speed": -1. + 2 * random.random()
                })

        elif event.type == 'Z' and event.value == 'PRESS':
            self.locate_z = True
            for obj in bpy.context.selected_objects:
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy(),
                    "speed": -1. + 2 * random.random()
                })

        elif event.type == 'LEFTMOUSE':
            self.scale = False
            self.rotate = False
            self.locate = False
            self.locate_z = False

        elif event.type in {'RET', 'ENTER', 'SPACE', 'NUMPAD_ENTER'}:
            MyHandleClass.remove_handler()
            self.scale = False
            self.rotate = False
            self.locate = False
            self.locate_z = False
            return {'FINISHED'}

        elif event.type == 'ESC':
            MyHandleClass.remove_handler()
            self.scale = False
            self.rotate = False
            self.locate = False
            self.locate_z = False

            for mesh in self.origin:
                obj = mesh["obj"]
                matrix = mesh["matrix"]
                obj.matrix_world = matrix

        elif event.type  == 'RIGHTMOUSE':
            MyHandleClass.remove_handler()
            self.scale = False
            self.rotate = False
            self.locate = False
            self.locate_z = False

            for mesh in self.origin:
                obj = mesh["obj"]
                matrix = mesh["matrix"]
                obj.matrix_world = matrix
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.object:
            self.first_mouse_x = event.mouse_x
            self.first_mouse_y = event.mouse_y
            self.meshes = []
            self.origin = []
            for obj in bpy.context.selected_objects:
                self.meshes.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy(),
                    "speed": -1. + 2 * random.random()
                })
                self.origin.append({
                    "obj": obj,
                    "matrix": obj.matrix_world.copy()
                })
            context.window_manager.modal_handler_add(self)

            MyHandleClass.remove_handler()
            MyHandleClass.add_handler(proflow_draw_callback_px, (x, y1, ps4, context.object))

            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "No active object, could not finish")
            return {'CANCELLED'}

# -----------------------------------------------------
# REGISTER/UNREGISTER
# -----------------------------------------------------
CLASSES =  [PROFLOW_OT_modal_cut,
            PROFLOW_OT_modal_extract,
            PROFLOW_OT_scale,
            PROFLOW_OT_modal_rotate_scale,
            PROFLOW_OT_loc_rot_scale_random
            ]

# Register
def register():
    for cls in CLASSES:
        try:
            bpy.utils.register_class(cls)
        except:
            print(f"{cls.__name__} already registred")

def unregister():

    for cls in reversed(CLASSES):
        try:
            bpy.utils.unregister_class(cls)
        except RuntimeError:
            pass
    # for cls in CLASSES:
    #     if hasattr(bpy.types, cls.__name__):
    #         bpy.utils.unregister_class(cls)

# if __name__ == "__main__":
#     register()


