# -*- coding:utf-8 -*-

# Proflow Add-on
# Copyright (C) 2018 Cedric Lepiller aka Pitiwazou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# <pep8 compliant>

import bpy
import bmesh
from bpy.types import Panel, Operator, PropertyGroup
from .preview import *
from .ui import *
import random
from bpy.props import (StringProperty,
                       BoolProperty,
                       PointerProperty,
                       FloatVectorProperty,
                       FloatProperty,
                       EnumProperty,
                       IntProperty,
                       BoolVectorProperty)
import math
from math import radians
from .functions import get_addon_preferences, recurLayerCollection,proflow_rename_collections
from mathutils import Matrix, Vector
import numpy as np
from random import uniform
# import time
# from mathutils import Matrix, Vector

def get_bmesh(ob):
    me = ob.data
    if ob.mode == 'OBJECT':
        bm = bmesh.new()
        bm.from_mesh(me)
        return bm

    bm = bmesh.from_edit_mesh(me)
    return bm

def finalize_bmesh(bm, ob):
    me = ob.data

    if ob.mode == 'OBJECT':
        bm.to_mesh(me)
        bm.free()
        del bm

    else:
        bmesh.update_edit_mesh(me)

def custom_select_deselect_all(bm, select=False):
    """ Select or deselect all
        select:
            - True = select all
            - False = deselect all """

    for v in bm.verts:
        v.select = select
        bm.select_flush(select)

def axis_change(self, context):
    proflow = context.window_manager.proflow

    # Check Direction
    # if context.object.get("x"):
    #     self.value = context.object["x"]
    # elif context.object.get("y"):
    #     self.value = context.object["y"]
    # elif context.object.get("z"):
    #     self.value = context.object["z"]
    # else:
    #     self.value = 1

    if context.object.get("proflow_array"):
        self.array_prop_value = context.object["proflow_array"]["value"]
        self.array_prop_axis = context.object["proflow_array"]["axis"]
        self.array_prop_merge = context.object["proflow_array"]["merge"]

        # proflow.directions = self.array_prop_axis
    # is_instance = False
    # if context.object.name.startswith("proflow_plane_instance"):
    #     is_instance = True

    curve = False
    array = False
    displace_1 = False
    displace_2 = False
    bend = False
    if context.object.modifiers:
        for mod in context.object.modifiers:
            if mod.type == 'CURVE':
                if mod.name == 'PFLOW - Curve':
                    self.mod_curve= mod
                    curve = True
            if mod.type == 'ARRAY':
                if mod.name == 'PFLOW - Array':
                    self.mod_array = mod
                    array = True
            if mod.type == 'DISPLACE':
                if mod.name == 'PFLOW - Displace - 1':
                    self.mod_displace_1 = mod
                    displace_1 = True
                if mod.name == 'PFLOW - Displace - 2':
                    self.mod_displace_2 = mod
                    displace_2 = True
            if mod.type == 'SIMPLE_DEFORM':
                if mod.name == "PFLOW - Bend":
                    self.mod_bend = mod
                    bend = True

    if self.directions == 'x':
        if curve:
            self.mod_curve.deform_axis = 'POS_X'
        if array:
            if context.object.get("proflow_array"):
                self.mod_array.relative_offset_displace[0] = self.array_prop_value
            else:
                self.mod_array.relative_offset_displace[0] = 1
            self.mod_array.relative_offset_displace[1] = 0
            self.mod_array.relative_offset_displace[2] = 0
        if bend:
            self.mod_bend.deform_axis = 'X'
        context.object.rotation_euler[1] = 0
        context.object.rotation_euler[2] = 0

    elif self.directions == 'y' :
        if curve:
            self.mod_curve.deform_axis = 'POS_Y'
        if array:
            self.mod_array.relative_offset_displace[0] = 0
            if context.object.get("proflow_array"):
                self.mod_array.relative_offset_displace[1] = self.array_prop_value
            else:
                self.mod_array.relative_offset_displace[1] = 1
            self.mod_array.relative_offset_displace[2] = 0
        if bend:
            self.mod_bend.deform_axis = 'Y'
        context.object.rotation_euler[0] = 0
        context.object.rotation_euler[2] = 0

    elif self.directions == 'z':
        if curve:
            self.mod_curve.deform_axis = 'POS_Z'
        if array:
            self.mod_array.relative_offset_displace[0] = 0
            self.mod_array.relative_offset_displace[1] = 0
            if context.object.get("proflow_array"):
                self.mod_array.relative_offset_displace[2] = self.array_prop_value
            else:
                self.mod_array.relative_offset_displace[2] = 1
        if bend:
            self.mod_bend.deform_axis = 'Z'
        context.object.rotation_euler[0] = 0
        context.object.rotation_euler[1] = 0

class PROFLOW_OT_rot_increments(Operator):
    '''
    Duplicate Edge and Add Asset on it
    '''
    bl_idname = 'object.proflow_rotation_increments'
    bl_label = ""
    bl_options = {'REGISTER'}

    rot_increments: EnumProperty(
        items=(('quarante_cinq', "quarante_cinq", ""),
               ('moins_quarante_cinq', "moins_quarante_cinq", ""),
               ('quatre_vingt_dix', "quatre_vingt_dix", ""),
               ('moins_quatre_vingt_dix', "moins_quatre_vingt_dix", ""),
               ('cent_quatre_vingt', "cent_quatre_vingt", ""),
               ),
        default='moins_quatre_vingt_dix')

    def execute(self, context):
        proflow = context.window_manager.proflow


        sel = [o for o in context.selected_objects]

        bpy.ops.object.select_all(action='DESELECT')

        for obj in sel:
            context.view_layer.objects.active = obj
            obj.select_set(state=True)
            if self.rot_increments == 'quarante_cinq':
                if proflow.directions == 'x':
                    obj.rotation_euler[0] += radians(45)
                elif proflow.directions == 'y':
                    obj.rotation_euler[1] += radians(45)
                elif proflow.directions == 'z':
                    obj.rotation_euler[2] += radians(45)

            elif self.rot_increments == 'moins_quarante_cinq':
                if proflow.directions == 'x':
                    obj.rotation_euler[0] += radians(-45)
                elif proflow.directions == 'y':
                    obj.rotation_euler[1] += radians(-45)
                elif proflow.directions == 'z':
                    obj.rotation_euler[2] += radians(-45)

            elif self.rot_increments == 'quatre_vingt_dix':
                if proflow.directions == 'x':
                    obj.rotation_euler[0] += radians(90)
                elif proflow.directions == 'y':
                    obj.rotation_euler[1] += radians(90)
                elif proflow.directions == 'z':
                    obj.rotation_euler[2] += radians(90)

            elif self.rot_increments == 'moins_quatre_vingt_dix':
                if proflow.directions == 'x':
                    obj.rotation_euler[0] += radians(-90)
                elif proflow.directions == 'y':
                    obj.rotation_euler[1] += radians(-90)
                elif proflow.directions == 'z':
                    obj.rotation_euler[2] += radians(-90)

            elif self.rot_increments == 'cent_quatre_vingt':
                if proflow.directions == 'x':
                    obj.rotation_euler[0] += radians(180)
                elif proflow.directions == 'y':
                    obj.rotation_euler[1] += radians(180)
                elif proflow.directions == 'z':
                    obj.rotation_euler[2] += radians(180)

        return {'FINISHED'}

class PROFLOW_OT_flow_loop(Operator):
    '''
    Duplicate Edge and Add Asset on it
    '''
    bl_idname = 'object.proflow_loop'
    bl_label = ""
    bl_options = {'REGISTER'}

    def execute(self, context):
        proflow = context.window_manager.proflow

        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "Do not work in Local Mode")
        #             return {'FINISHED'}

        # Check if Custom mesh is instance
        if proflow.custom_mesh is None:
            proflow.use_custom_mesh = False
        if proflow.use_custom_mesh:
            if proflow.custom_mesh.type == 'EMPTY':
                proflow.use_plane_instance = True

        act_obj = context.active_object
        # self.smooth_verts = False

        proflow.active_mesh = act_obj
        addon_pref = get_addon_preferences()

        # Material et mesh
        # if act_obj.type == 'MESH' and len(act_obj.material_slots) and addon_pref.copy_material:
        if len(act_obj.material_slots) and addon_pref.copy_material:
            self.mat = act_obj.active_material
            proflow.act_obj_material_name = self.mat.name
        else:
            proflow.act_obj_material_name = ""
            addon_pref.copy_material = False
            proflow.replace_asset = False

        # Si on a des polys sinon on passe
        if len(context.object.data.polygons):
            ob = context.object
            me = ob.data
            bm = bmesh.from_edit_mesh(me)
            if any([e.select for e in bm.edges]):
                # act_obj = context.active_object

                # On Check si on a plus de 20 vertex
                # C = bpy.context
                # ob = C.object
                # bm = get_bmesh(ob)
                # if len(bm.verts) >= 20:
                #     self.smooth_verts = True

                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.duplicate_move()
                bpy.ops.object.mode_set(mode='EDIT')

                if tuple(context.tool_settings.mesh_select_mode) == (False, False, True):
                    bpy.ops.mesh.select_all(action='INVERT')
                    bpy.ops.mesh.delete(type='FACE')
                else:
                    bpy.ops.mesh.select_mode(use_extend = False, use_expand=False, type='VERT')
                    bpy.ops.mesh.select_all(action = 'INVERT')
                    bpy.ops.mesh.delete(type = 'VERT')
                bpy.ops.mesh.select_all(action = 'SELECT')
                bpy.ops.object.mode_set(mode = 'OBJECT')
        # else:
        #     addon_pref.copy_material = False

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
        bpy.ops.object.convert(target = 'CURVE')

        # print("PROUUUUUT")
        if proflow.act_obj_material_name is not None and addon_pref.copy_material:
            existing_mat = bpy.data.materials[proflow.act_obj_material_name]
            context.object.active_material = bpy.data.materials[existing_mat.name]

        bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')

        bpy.ops.object.proflow_finalize()

        return {'FINISHED'}

class PROFLOW_OT_draw_flow(Operator):
    '''
    DRAW FLOATER

    This tool will allow you to draw on selected OBJECT or in the 3Dview.

    SHIFT - Use Selected Object
    '''
    bl_idname = 'object.proflow_draw'
    bl_label = ""
    bl_options = {'REGISTER'}

    def invoke(self, context, event):
        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "Do not work in Local Mode")
        #             return {'FINISHED'}

        proflow = context.window_manager.proflow

        # Check if Custom mesh is instance
        if proflow.custom_mesh is None:
            proflow.use_custom_mesh = False
        if proflow.use_custom_mesh:
            if proflow.custom_mesh.type == 'EMPTY':
                proflow.use_plane_instance = True

        proflow.draw = True
        addon_pref = get_addon_preferences()

        act_obj = context.active_object
        proflow.active_mesh = act_obj
        proflow.no_material = False

        if event.shift:
            proflow.use_selected_custom_mesh = True
            proflow.use_custom_mesh = True
            proflow.custom_mesh = act_obj

        # Material
        if act_obj:
            # if act_obj.type == 'MESH' and len(act_obj.material_slots) and addon_pref.copy_material:
            if len(act_obj.material_slots) and addon_pref.copy_material:
                self.mat = act_obj.active_material
                proflow.act_obj_material_name = self.mat.name
            else:
                proflow.act_obj_material_name = ""
                proflow.replace_asset = False
                proflow.no_material = True

            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.select_all(action='DESELECT')
        else:
            proflow.no_material = True

        # Clear Cursor Loc & Rot
        context.scene.cursor.location[:] = (0,0,0)
        context.scene.cursor.rotation_euler[:] = (0,0,0)

        # Curve creation
        bpy.ops.curve.primitive_bezier_curve_add(radius=1, enter_editmode=True, align='WORLD')

        self.new_curve = context.object
        context.space_data.overlay.show_curve_normals = False
        bpy.ops.curve.delete(type='VERT')
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

        bpy.ops.object.select_all(action='DESELECT')

        bpy.ops.object.proflow_finalize()


        # On sort avec une courbe comme selection
        return {'FINISHED'}

class PROFLOW_OT_from_boolean(Operator):
    '''
    CREATE FLOATER

    The FLOATER will be created on selection intersection.
    You need to select 2 or 3 objects.

    SHIFT - Extract Edges
    '''
    bl_idname = 'object.proflow_from_boolean'
    bl_label = ""
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        return True

    def invoke(self, context, event):
        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "Do not work in Local Mode")
        #             return {'FINISHED'}

        proflow = context.window_manager.proflow
        addon_pref = get_addon_preferences()
        proflow.from_booleans = True

        # Check if Custom mesh is instance
        if proflow.custom_mesh is None:
            proflow.use_custom_mesh = False
        if proflow.use_custom_mesh:
            if proflow.custom_mesh.type == 'EMPTY':
                proflow.use_plane_instance = True

        first_act_obj = context.active_object
        if not proflow.replace_asset:
            proflow.active_mesh = first_act_obj
            if len(context.selected_objects) == 2:
                proflow.auxiliary_mesh = context.selected_objects[0] if context.selected_objects[0] != first_act_obj else \
                context.selected_objects[1]
        # print("ACTIVE MESH: ", proflow.active_mesh)

        # smooth_verts = False
        bpy.ops.object.duplicate_move()

        act_obj = context.active_object
        # proflow.active_mesh = act_obj
        # print("ACTIVE MESH: ",proflow.active_mesh)

        self.exract_edges = False

        if event.shift:
            self.exract_edges = True

        bpy.ops.object.convert(target='MESH')

        # Material
        # if act_obj.type == 'MESH' and len(act_obj.material_slots) and addon_pref.copy_material:
        if len(act_obj.material_slots) and addon_pref.copy_material:
            self.mat = act_obj.active_material
            proflow.act_obj_material_name = self.mat.name
        else:
            proflow.act_obj_material_name = ""
            addon_pref.copy_material = False
            proflow.replace_asset = False

        if len([obj for obj in context.selected_objects ]) > 2:
            Union_Objects(context)

        act_obj = context.active_object
        obj1, obj2 = context.selected_objects
        sel_obj = obj1 if obj1 != act_obj else obj2

        if len([obj for obj in context.selected_objects if context.object is not None if obj.type =='MESH']) > 2:
            act_obj.select_set(state=False)
            for obj in context.selected_objects:
                if obj.modifiers:
                    context.view_layer.objects.active = obj
                    for mod in obj.modifiers:
                        if mod.show_viewport == False:
                            bpy.ops.object.modifier_remove(modifier=mod.name)
                        else:
                            if bpy.app.version >= (2, 90, 0):
                                bpy.ops.object.modifier_apply(modifier=mod.name)
                            else:
                                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=mod.name)

        # On deselectionne les faces de l'objet actif
        C = context
        ob = C.object
        bm = get_bmesh(ob)
        custom_select_deselect_all(bm, select=False)
        finalize_bmesh(bm, ob)

        newMod = act_obj.modifiers.new("Boolean", "BOOLEAN")
        newMod.object = sel_obj
        newMod.operation = 'DIFFERENCE'
        act_obj.select_set(state=False)

        # On selectione les edges du selobj
        context.view_layer.objects.active = sel_obj
        sel_obj.select_set(state=True)
        C = context
        ob = C.object
        bm = get_bmesh(ob)
        custom_select_deselect_all(bm, select=True)
        finalize_bmesh(bm, ob)
        sel_obj.select_set(state=False)

        # on reselectionn l'obj actif et on applique le modifier
        context.view_layer.objects.active = act_obj
        act_obj.select_set(state=True)
        if bpy.app.version >= (2, 90, 0):
            bpy.ops.object.modifier_apply(modifier="Boolean")
        else:
            bpy.ops.object.modifier_apply(apply_as='DATA', modifier="Boolean")

        # on selectionne les boundary du boolean et on cree l'objet
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.region_to_loop()

        #On Check si on a plus de 20 vertex
        # C = context
        # ob = C.object
        # bm = get_bmesh(ob)
        # if len(bm.verts)>= 20:
        #     smooth_verts = True

        bpy.ops.mesh.duplicate_move()
        bpy.ops.mesh.separate(type='SELECTED')
        bpy.ops.object.mode_set(mode='OBJECT')
        act_obj.select_set(state=False)

        # On supprime les copies
        bpy.data.objects.remove(act_obj, do_unlink=True)
        bpy.data.objects.remove(sel_obj, do_unlink=True)

        # On passe la selection en active
        context.view_layer.objects.active = context.selected_objects[0]
        bpy.ops.object.mode_set(mode='OBJECT')

        if not self.exract_edges:

            # On nettoie le mesh et convertit en curve
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
            bpy.ops.mesh.select_all(action='SELECT')

            # delete la seconde courbe
            bpy.ops.mesh.separate(type='LOOSE')
            bpy.ops.object.mode_set(mode='OBJECT')

            self.sel_curves = []
            several_objs = False
            if len(context.selected_objects) > 1:
                several_objs = True
            if several_objs:
                act_obj = context.active_object
                for obj in context.selected_objects:
                    if obj == act_obj:
                        continue
                    self.sel_curves.append(obj)
                    obj.select_set(state=False)
                for obj in self.sel_curves:
                    bpy.data.objects.remove(obj, do_unlink=True)

            bpy.ops.object.convert(target='CURVE')
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            bpy.ops.object.proflow_finalize()

            if several_objs:
                del (self.sel_curves[:])
        # On sort avec une courbe comme selection
        return {'FINISHED'}

def local_rotate(obj, axis, angle):
    obj.matrix_world @= Matrix.Rotation(math.radians(angle), 4, axis)

class PROFLOW_OT_replace_asset(Operator):
    '''
    REPLACE ASSET

    Replace the current asset by the new one selected from the preview or from a custom asset
    '''
    bl_idname = "object.proflow_replace_asset"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        # for area in bpy.context.screen.areas:
        #     if area.type == 'VIEW_3D':
        #         if area.spaces.active.local_view:
        #             self.report({'WARNING'}, "Do not work in Local Mode")
        #             return {'FINISHED'}

        proflow = context.window_manager.proflow

        if len(context.selected_objects) == 0:
            self.report({'WARNING'}, "You need to select one object!")
            return {'FINISHED'}


        # Check if Custom mesh is instance
        if proflow.custom_mesh is None:
            proflow.use_custom_mesh = False
        if proflow.use_custom_mesh:
            if proflow.custom_mesh.type == 'EMPTY':
                proflow.use_plane_instance = True


        addon_pref = get_addon_preferences()
        current_asset = context.active_object
        # copy rotation
        self.current_mesh_matrix_save = current_asset.matrix_world.copy()

        bpy.ops.view3d.snap_cursor_to_selected()
        self.replace_profile = False

        # if not current_asset.material_slots:
        #     addon_pref.copy_material = False

        # on recup la courbe
        if current_asset.type in {'MESH', 'FONT'} and current_asset.modifiers :
            for mod in current_asset.modifiers:
                if mod.type == 'CURVE' and mod.name == 'PFLOW - Curve':
                    self.curve = mod.object
                # On recup le mesh actif pour shrinkwrap et data transfert bpy.ops.object.modifier_add(type='DATA_TRANSFER')
                if mod.type == 'SHRINKWRAP':
                    if mod.target is not None:
                        proflow.active_mesh = mod.target

                    if mod.auxiliary_target is not None:
                        proflow.auxiliary_mesh = mod.auxiliary_target

        # on recup la courbe et le profile et on clean le profile
        elif current_asset.type == 'CURVE' and current_asset.data.bevel_object is not None:
            self.replace_profile = True
            self.current_profile = current_asset.data.bevel_object
            self.curve = current_asset
            # self.profile = context.object.data.bevel_object
            self.curve.data.bevel_object = None
            context.object.data.bevel_depth = 0

        # Material
        if len(self.curve.material_slots) and addon_pref.copy_material:
            self.mat = self.curve.active_material
            proflow.act_obj_material_name = self.mat.name
        else:
            proflow.act_obj_material_name = ""
            addon_pref.copy_material = False
            proflow.replace_asset = False

        # on import le mesh de la biblio ou celui de la scene
        if not proflow.use_custom_mesh:
            for area in bpy.context.screen.areas:
                if area.type == 'VIEW_3D':
                    if area.spaces.active.local_view:
                        proflow.active_coll = context.view_layer.active_layer_collection.name
                        layer_collection = context.view_layer.layer_collection
                        layerColl = recurLayerCollection(layer_collection, 'Proflow')
                        context.view_layer.active_layer_collection = layerColl

            bpy.ops.object.proflow_assets_add()

        else:
            bpy.ops.object.select_all(action='DESELECT')
            for obj in context.scene.objects:
                if obj.name == proflow.custom_mesh.name:
                    context.view_layer.objects.active = obj
                    obj.select_set(state=True)
                    bpy.ops.object.duplicate_move()

        proflow.replace_asset_mesh = context.active_object

        if context.scene.unit_settings.length_unit == 'METERS':
            context.object.dimensions = context.object.dimensions / 10
            bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        elif context.scene.unit_settings.length_unit == 'MILLIMETERS':
            context.object.dimensions = context.object.dimensions * 10
            bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

        proflow.replace_asset_mesh.matrix_world = self.current_mesh_matrix_save
        # new object from copy custom asset
        if not self.replace_profile :
            bpy.data.objects.remove(current_asset, do_unlink=True)
        else:
            bpy.data.objects.remove(self.current_profile, do_unlink=True)

        if proflow.replace_asset_mesh.type == 'CURVE':
            self.replace_profile = True
        elif proflow.replace_asset_mesh.type == 'MESH':
            self.replace_profile = False

        #on reselect la courbe
        bpy.ops.object.select_all(action='DESELECT')
        context.view_layer.objects.active = self.curve
        self.curve.select_set(state=True)

        if self.replace_profile:
            self.curve.data.bevel_object = proflow.replace_asset_mesh
            bpy.ops.object.proflow_modal_rotate_scale('INVOKE_DEFAULT')
            return {'FINISHED'}
        else:
            bpy.ops.object.proflow_finalize()

        # On sors avec une courbe
        return {'FINISHED'}

class PROFLOW_OT_finalize(Operator):
    bl_idname = 'object.proflow_finalize'
    bl_label = ""
    bl_options = {'REGISTER'}

    # -------------------------------------
    # COLLECTIONS
    # -------------------------------------

    def create_collections(self, context):
        collection_name = "Proflow"
        myCol = bpy.data.collections.new(collection_name)
        context.scene.collection.children.link(myCol)


    def move_to_collection(self, obj, name):
        keys = bpy.context.view_layer.layer_collection.children.keys()

        if not name in keys:
            bpy.ops.object.move_to_collection(collection_index=0, is_new=True, new_collection_name=name)
        else:
            for index, key in enumerate(keys):
                if key == name:
                    bpy.ops.object.move_to_collection(collection_index=index + 1)
                    break


    # def rename_collections(self, context):
    #     # Rename collection
    #     keys = list(context.scene.collection.children.keys())
    #     if len(bpy.data.scenes) >= 2:
    #         for key in keys:
    #             coll = bpy.data.collections[key]
    #             if coll.name.startswith("Proflow"):
    #                 coll.name = "Proflow_" + context.scene.name
    #                 self.collection_visibility = coll
    #                 self.coll_is_hidden = False
    #                 if context.view_layer.layer_collection.children[self.collection_visibility.name].exclude:
    #                     self.coll_is_hidden = True
    #                     context.view_layer.layer_collection.children[
    #                         self.collection_visibility.name].exclude = False
    #
    #     elif len(bpy.data.scenes) == 1:
    #         for key in keys:
    #             coll = bpy.data.collections[key]
    #             if coll.name.startswith("Proflow"):
    #                 coll.name = "Proflow"
    #                 # bpy.context.scene.exclude = False
    #
    #                 self.collection_visibility = coll
    #                 self.coll_is_hidden = False
    #                 if context.view_layer.layer_collection.children[self.collection_visibility.name].exclude:
    #                     self.coll_is_hidden = True
    #                     context.view_layer.layer_collection.children[
    #                         self.collection_visibility.name].exclude = False

    def move_active_object_to_collection(self, context):
        if len(bpy.data.scenes) >= 2:
            self.proflow_collection = "Proflow_" + context.scene.name
            keys = list(context.scene.collection.children.keys())
            if len(bpy.data.scenes) >= 2:
                for key in keys:
                    coll = bpy.data.collections[key]
                    if coll.name.startswith("Proflow"):
                        if context.view_layer.layer_collection.children[self.proflow_collection].exclude:
                            context.view_layer.layer_collection.children[self.proflow_collection].exclude = False
            self.move_to_collection(context.active_object, self.proflow_collection)
        elif len(bpy.data.scenes) == 1:
            self.proflow_collection = "Proflow"
            keys = list(context.scene.collection.children.keys())
            if len(bpy.data.scenes) >= 2:
                for key in keys:
                    coll = bpy.data.collections[key]
                    if coll.name.startswith("Proflow"):
                        if context.view_layer.layer_collection.children[self.proflow_collection].exclude:
                            context.view_layer.layer_collection.children[self.proflow_collection].exclude = False
            self.move_to_collection(context.active_object, self.proflow_collection)



    on_curve: BoolProperty(name='', default=False, description='On Curve')

    def execute(self, context):
        proflow = context.window_manager.proflow
        addon_pref = get_addon_preferences()
        self.set_profile = False

        if self.on_curve:
            proflow.replace_asset = False
            addon_pref.copy_material = False

        # get proflow collection
        keys = list(bpy.context.scene.collection.children.keys())
        for key in keys:
            coll = bpy.data.collections[key]
            if coll.name.startswith("Proflow"):
                self.proflow_collection = coll

        proflow_rename_collections(self, context)

        proflow.active_coll = context.view_layer.active_layer_collection.name
        if addon_pref.add_to_collection:
            if len(bpy.data.scenes) >= 2:
                self.proflow_collection = "Proflow_" + context.scene.name  # on store le nom
            else:
                self.proflow_collection = "Proflow"

            pf_collection = context.view_layer.layer_collection.children[self.proflow_collection]
            context.view_layer.active_layer_collection = pf_collection  # pour metre la collection active

        if addon_pref.add_to_collection:
            for area in bpy.context.screen.areas:
                if area.type == 'VIEW_3D':
                    if area.spaces.active.local_view:
                        if proflow.active_coll.strip() == "":
                            proflow.active_coll = context.view_layer.active_layer_collection.name
                            layer_collection = context.view_layer.layer_collection
                            layerColl = recurLayerCollection(layer_collection, 'Proflow')
                            context.view_layer.active_layer_collection = layerColl
                    else:
                        self.move_active_object_to_collection(context)

        self.curve_obj = context.active_object

        if proflow.replace_asset:
            bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')
            bpy.ops.view3d.snap_cursor_to_selected()
            bpy.ops.object.select_all(action='DESELECT')
            self.mesh_obj = proflow.replace_asset_mesh
            context.view_layer.objects.active = self.mesh_obj
            self.mesh_obj.select_set(state=True)
            bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)
        else:
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.curve.select_all(action='SELECT')
            context.space_data.overlay.show_curve_normals = False
            bpy.ops.curve.spline_type_set(type='BEZIER')
            if proflow.from_booleans:
                bpy.ops.curve.handle_type_set(type='VECTOR')
            else:
                bpy.ops.curve.handle_type_set(type='AUTOMATIC')

            if proflow.from_booleans:
                bpy.ops.transform.tilt(value=-0.785398, mirror=True, use_proportional_edit=False,
                                       proportional_edit_falloff='SMOOTH', proportional_size=1,
                                       use_proportional_connected=False, use_proportional_projected=False)

            bpy.ops.object.mode_set(mode='OBJECT')
            context.object.show_in_front = False
            bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')
            bpy.ops.view3d.snap_cursor_to_selected()
            bpy.ops.object.select_all(action='DESELECT')
            if not proflow.use_custom_mesh:
                bpy.ops.object.proflow_assets_add()

            else:
                for obj in context.scene.objects:
                    if obj.name == proflow.custom_mesh.name:
                        context.view_layer.objects.active = obj
                        obj.select_set(state=True)

                        # Duplique les start et end cap
                        for mod in obj.modifiers:
                            if mod.type == 'ARRAY':
                                if mod.start_cap is not None:
                                    self.array_start_cap = mod.start_cap
                                    self.array_start_cap.select_set(state=True)
                                if mod.end_cap is not None:
                                    self.array_end_cap = mod.end_cap
                                    self.array_end_cap.select_set(state=True)

                                bpy.ops.object.duplicate_move()
                                active_object = context.active_object
                                bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)

                                for obj in context.selected_objects:
                                    if obj == active_object:
                                        continue
                                    obj.hide_set(True)

                                bpy.ops.object.select_all(action='DESELECT')
                                context.view_layer.objects.active = active_object

                            else:
                                bpy.ops.object.duplicate_move()
                                bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)



                        new_mesh = context.active_object
                        proflow.custom_mesh.name = new_mesh.name


            self.mesh_obj = context.active_object


            if context.scene.unit_settings.length_unit == 'METERS':
                context.object.dimensions = context.object.dimensions / 10
                bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
            elif context.scene.unit_settings.length_unit == 'MILLIMETERS':
                context.object.dimensions = context.object.dimensions * 10
                bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)


            # bpy.ops.object.select_grouped(type='PARENT')

            # if proflow.make_booleans:
            #     if proflow.active_mesh.type == 'MESH' and proflow.active_mesh.display_type == 'BOUNDS':
            #         boolean_modifier = proflow.active_mesh.modifiers.new('Boolean', 'BOOLEAN')
            #         boolean_modifier.object = self.mesh_obj

            if self.mesh_obj.type == 'CURVE':
                self.set_profile = True

            # Check for Plane instance
            elif self.mesh_obj.type == 'MESH':
                if self.mesh_obj.get("proflow_plane_instance"):
                    proflow.use_plane_instance = True

        bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)

        # Set bevel profile
        if self.set_profile:
            self.curve_obj.data.bevel_mode = 'OBJECT'
            self.curve_obj.data.bevel_object = self.mesh_obj
            # self.curve_obj.data.splines[0].use_smooth = False
            # self.curve_obj.data.splines[0].resolution_u = 1
            bpy.ops.object.select_all(action='DESELECT')
            context.view_layer.objects.active = self.curve_obj
            self.curve_obj.select_set(state=True)
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.curve.switch_direction()
            bpy.ops.object.mode_set(mode='OBJECT')
        else:
            if proflow.use_plane_instance:
                bpy.ops.mesh.primitive_plane_add(size=0.1)
                self.plane_obj = context.active_object
                self.plane_obj.name = "proflow_plane_instance"

                if self.mesh_obj.get("proflow_array"):
                    self.array_prop_value = self.mesh_obj["proflow_array"]["value"]
                    self.array_prop_axis = self.mesh_obj["proflow_array"]["axis"]
                    self.array_prop_merge = self.mesh_obj["proflow_array"]["merge"]
                    proflow.directions = self.array_prop_axis.lower()
                    self.plane_obj["proflow_array"] = {"axis": self.array_prop_axis, "value": self.array_prop_value,
                                           "merge": self.array_prop_merge}


                self.mesh_obj.select_set(state=True)
                context.view_layer.objects.active = self.plane_obj
                bpy.ops.object.parent_set(type='OBJECT')
                # bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)

                self.mesh_obj.select_set(state=False)
                if self.mesh_obj == 'MESH' :
                    bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
                # if self.mesh_obj == 'MESH' and not self.mesh_obj.children:
                #     bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

                self.plane_obj.instance_type = 'FACES'
                context.object.show_instancer_for_render = False
                self.plane_obj["proflow_plane_instance"] = True

                bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)


            # CHECK ARRAY
            if self.mesh_obj.get("proflow_array"):
                self.array_prop_value = self.mesh_obj["proflow_array"]["value"]
                self.array_prop_axis = self.mesh_obj["proflow_array"]["axis"]
                self.array_prop_merge = self.mesh_obj["proflow_array"]["merge"]
                proflow.directions = self.array_prop_axis

            # CHECK SUBSURF
            self.add_subsurf = False
            if self.mesh_obj.get("proflow_subsurf"):
                self.add_subsurf = True
                self.subsurf_value = self.mesh_obj["proflow_subsurf"]

            # CHECK BEND
            self.add_bend = False
            if self.mesh_obj.get("proflow_bend"):
                self.add_bend = True
                self.bend_value = self.mesh_obj["proflow_bend"]["value"]
                self.bend_axis = self.mesh_obj["proflow_bend"]["axis"]

            # CHECK WELD
            self.add_weld = False
            if self.mesh_obj.get("proflow_weld"):
                self.add_weld = True

            # CHECK DATA
            self.add_data = False
            if self.mesh_obj.get("proflow_data_transfer"):
                self.add_data = True

            self.use_existing_array = False
            if self.mesh_obj.modifiers:
                for mod in self.mesh_obj.modifiers:
                    if mod.type == 'CURVE':
                        if mod.name == 'PFLOW - Curve':
                            bpy.ops.object.modifier_remove(modifier="PFLOW - Curve")
                    if mod.type == 'ARRAY':
                        if mod.fit_type == 'FIT_CURVE':
                            self.use_existing_array = True
                        elif mod.name == 'PFLOW - Array':
                            bpy.ops.object.modifier_remove(modifier="PFLOW - Array")
                    if mod.type == 'WELD':
                        if mod.name == 'PFLOW - Weld':
                            bpy.ops.object.modifier_remove(modifier="PFLOW - Weld")
                    if mod.type == 'SUBSURF':
                        if mod.name == 'Subdivision':
                            bpy.ops.object.modifier_remove(modifier="Subdivision")
                    if mod.type == 'DISPLACE':
                        if mod.name == 'PFLOW - Displace - 1':
                            bpy.ops.object.modifier_remove(modifier="PFLOW - Displace - 1")
                        if mod.name == 'PFLOW - Displace - 2':
                            bpy.ops.object.modifier_remove(modifier="PFLOW - Displace - 2")
                    if mod.type == 'SIMPLE_DEFORM':
                        if mod.name == "PFLOW - SimpleDeform":
                            bpy.ops.object.modifier_remove(modifier="PFLOW - SimpleDeform")

            if not proflow.use_plane_instance:
                # ADD SIMPLE_DEFORM
                if self.add_bend:
                    bend = self.mesh_obj.modifiers.new('PFLOW - Bend', 'SIMPLE_DEFORM')
                    bend.deform_method = 'BEND'
                    if self.add_bend:
                        bend.angle = self.bend_value
                        bend.deform_axis = self.bend_axis.upper()
                    else:
                        bend.angle = radians(0.01)
                        bend.deform_axis = 'Z'
                    bend.show_viewport = True
                    bend.show_expanded = False
                    bend.show_in_editmode = True
                    bend.show_on_cage = True

            # Add DISPLACE Y
            if proflow.use_plane_instance:
                displace_modifier_y = self.plane_obj.modifiers.new('PFLOW - Displace - 2', 'DISPLACE')
            else:
                displace_modifier_y = self.mesh_obj.modifiers.new('PFLOW - Displace - 2', 'DISPLACE')
            if proflow.directions == 'x':
                displace_modifier_y.direction = 'Z'
            elif proflow.directions == 'y':
                displace_modifier_y.direction = 'Z'
            elif proflow.directions == 'z':
                displace_modifier_y.direction = 'Y'
            displace_modifier_y.mid_level = 0
            displace_modifier_y.strength = 0.0001
            displace_modifier_y.show_viewport = True
            displace_modifier_y.show_expanded = False
            displace_modifier_y.show_in_editmode = True
            displace_modifier_y.show_on_cage = True

            # Add DISPLACE X
            if proflow.use_plane_instance:
                displace_modifier_x = self.plane_obj.modifiers.new('PFLOW - Displace - 1', 'DISPLACE')
            else:
                displace_modifier_x = self.mesh_obj.modifiers.new('PFLOW - Displace - 1', 'DISPLACE')
            displace_modifier_x.direction = proflow.directions.upper()
            displace_modifier_x.mid_level = 0
            displace_modifier_x.strength = 0.0001
            displace_modifier_x.show_viewport = True
            displace_modifier_x.show_expanded = False
            displace_modifier_x.show_in_editmode = True
            displace_modifier_x.show_on_cage = True

            # Add ARRAY
            if not self.use_existing_array :
                if proflow.use_plane_instance:
                    self.array_modifier = self.plane_obj.modifiers.new('PFLOW - Array', 'ARRAY')
                else:
                    self.array_modifier = self.mesh_obj.modifiers.new('PFLOW - Array', 'ARRAY')
            else:
                for mod in self.mesh_obj.modifiers:
                    if mod.type == 'ARRAY':
                        mod.name = 'PFLOW - Array'
                        self.array_modifier = mod

            self.array_modifier.fit_type = 'FIT_CURVE'
            self.array_modifier.show_viewport = True
            self.array_modifier.show_expanded = False
            self.array_modifier.show_in_editmode = True
            self.array_modifier.show_on_cage = True

            if self.mesh_obj.get("proflow_array") and not proflow.use_plane_instance:
                self.array_modifier.use_merge_vertices = self.array_prop_merge
            else:
                self.array_modifier.use_merge_vertices = False

            if self.array_modifier.use_merge_vertices:
                self.array_modifier.merge_threshold = 0.001

            self.array_modifier.curve = self.curve_obj

            if self.mesh_obj.get("proflow_array"):
                if proflow.directions == 'x':
                    if proflow.use_plane_instance:
                        self.array_modifier.relative_offset_displace[0] = self.array_prop_value * 4
                    else:
                        self.array_modifier.relative_offset_displace[0] = self.array_prop_value
                    self.array_modifier.relative_offset_displace[1] = 0
                    self.array_modifier.relative_offset_displace[2] = 0
                elif proflow.directions == 'y':
                    self.array_modifier.relative_offset_displace[0] = 0
                    if proflow.use_plane_instance:
                        self.array_modifier.relative_offset_displace[1] = self.array_prop_value * 4
                    else:
                        self.array_modifier.relative_offset_displace[1] = self.array_prop_value
                    self.array_modifier.relative_offset_displace[2] = 0
                elif proflow.directions == 'z':
                    self.array_modifier.relative_offset_displace[0] = 0
                    self.array_modifier.relative_offset_displace[1] = 0
                    if proflow.use_plane_instance:
                        self.array_modifier.relative_offset_displace[2] = self.array_prop_value * 4
                    else:
                        self.array_modifier.relative_offset_displace[2] = self.array_prop_value
            else:
                self.array_modifier.relative_offset_displace[0] = 0
                self.array_modifier.relative_offset_displace[1] = 0
                if proflow.use_plane_instance:
                    self.array_modifier.relative_offset_displace[2] = 4
                else:
                    self.array_modifier.relative_offset_displace[2] = 1

            # ADD CURVE MODIFER
            if proflow.use_plane_instance:
                curve_modifier = self.plane_obj.modifiers.new('PFLOW - Curve', 'CURVE')
            else:
                curve_modifier = self.mesh_obj.modifiers.new('PFLOW - Curve', 'CURVE')
            curve_modifier.object = self.curve_obj
            curve_modifier.show_viewport = True
            curve_modifier.show_expanded = False
            curve_modifier.show_in_editmode = True
            curve_modifier.show_on_cage = True
            if proflow.directions == 'x':
                curve_modifier.deform_axis = 'POS_X'
            elif proflow.directions == 'y':
                curve_modifier.deform_axis = 'POS_Y'
            elif proflow.directions == 'z':
                curve_modifier.deform_axis = 'POS_Z'
            if self.add_weld and not self.add_data:
                bpy.ops.object.modifier_move_down(modifier="PFLOW - Curve")

            # ADD WELD
            if self.add_weld:
                weld_modifier = self.mesh_obj.modifiers.new('PFLOW - Weld', 'WELD')
                weld_modifier.merge_threshold = 0
                unit = context.scene.unit_settings.scale_length
                if np.isclose(1, unit):
                    weld_modifier.merge_threshold = 0.01
                elif np.isclose(0.01, unit):
                    weld_modifier.merge_threshold = 0.001
                elif np.isclose(0.001, unit):
                    weld_modifier.merge_threshold = 0.0001
                weld_modifier.show_in_editmode = True
                weld_modifier.show_on_cage = True
                weld_modifier.show_expanded = False

                if self.mesh_obj.vertex_groups:
                    for vgroups in self.mesh_obj.vertex_groups:
                        if vgroups.name.startswith("pf_group_w"):
                            weld_modifier.vertex_group = "pf_group_w"

            # ADD SUBSURF
            if self.add_subsurf:

                subsurf_modifier = self.mesh_obj.modifiers.new('Subdivision', 'SUBSURF')
                subsurf_modifier.levels = self.subsurf_value
                # if not proflow.edit_with_modal:
                subsurf_modifier.show_viewport = False
                # else:
                #     subsurf_modifier.show_viewport = False
                subsurf_modifier.show_expanded = False
                subsurf_modifier.show_in_editmode = True
                subsurf_modifier.show_on_cage = True

            # print("AUXILIAR MESH: ", proflow.auxiliary_mesh.name)
            # TRANSFERT NORMALS
            # if proflow.Transfert_normals:
            if self.add_data:
                # context.object.data.use_auto_smooth = True


                has_vertex_groups = False

                for vgroup in self.mesh_obj.vertex_groups:
                    if vgroup.name.startswith("pf_group_"):
                        has_vertex_groups = True

                if has_vertex_groups:
                    new_shrinkwarp = self.mesh_obj.modifiers.new(name='PFLOW - Shrinkwrap', type='SHRINKWRAP')
                    # if proflow.replace_asset and proflow.target_mesh is not None:
                    #     new_shrinkwarp.target = proflow.target_mesh
                    # else:
                    new_shrinkwarp.target = proflow.active_mesh
                    new_shrinkwarp.wrap_method = 'PROJECT'
                    new_shrinkwarp.cull_face = 'OFF'
                    # if proflow.auxiliary_mesh:
                    new_shrinkwarp.use_negative_direction = True
                    # else:
                    #     new_shrinkwarp.use_negative_direction = False
                    new_shrinkwarp.use_positive_direction = True

                    if proflow.auxiliary_mesh:
                        new_shrinkwarp.auxiliary_target = proflow.auxiliary_mesh
                    new_shrinkwarp.vertex_group = "pf_group_0"
                    new_shrinkwarp.show_expanded = False

                    if not proflow.auxiliary_mesh:
                        new_data_transfert = self.mesh_obj.modifiers.new(name='PFLOW - DataTransfer', type='DATA_TRANSFER')
                        # if proflow.replace_asset and proflow.target_mesh is not None:
                        #     new_data_transfert.object = proflow.target_mesh
                        # else:
                        new_data_transfert.object = proflow.active_mesh
                        new_data_transfert.vertex_group = "pf_group_0"
                        new_data_transfert.use_loop_data = True
                        new_data_transfert.data_types_loops = {'CUSTOM_NORMAL'}
                        new_data_transfert.loop_mapping = 'POLYINTERP_LNORPROJ'
                        new_data_transfert.show_expanded = False
                    else:
                        # bpy.context.object.data.auto_smooth_angle = radians(180)

                        new_data_transfert_1 = self.mesh_obj.modifiers.new(name='PFLOW - DataTransfer_1', type='DATA_TRANSFER')
                        # if proflow.replace_asset and proflow.target_mesh is not None:
                        #     new_data_transfert_1.object = proflow.target_mesh
                        # else:
                        new_data_transfert_1.object = proflow.active_mesh
                        new_data_transfert_1.vertex_group = "pf_group_1"
                        new_data_transfert_1.use_loop_data = True
                        new_data_transfert_1.data_types_loops = {'CUSTOM_NORMAL'}
                        new_data_transfert_1.loop_mapping = 'POLYINTERP_LNORPROJ'
                        new_data_transfert_1.show_expanded = False

                        new_data_transfert_2 = self.mesh_obj.modifiers.new(name='PFLOW - DataTransfer_2',type='DATA_TRANSFER')
                        new_data_transfert_2.object = proflow.auxiliary_mesh
                        new_data_transfert_2.vertex_group = "pf_group_2"
                        new_data_transfert_2.use_loop_data = True
                        new_data_transfert_2.data_types_loops = {'CUSTOM_NORMAL'}
                        new_data_transfert_2.loop_mapping = 'POLYINTERP_LNORPROJ'
                        new_data_transfert_2.show_expanded = False

        # if not proflow.use_plane_instance:
        if addon_pref.copy_material:
            if not proflow.no_material :
                existing_mat = bpy.data.materials[proflow.act_obj_material_name]
                if self.mesh_obj.name.startswith("proflow_plane_instance"):
                    # print("YES")
                    if self.mesh_obj.children:
                        bpy.ops.object.select_grouped(type='CHILDREN')
                        child = context.select_objects
                        context.view_layer.objects.active = child
                        child.active_material = bpy.data.materials[existing_mat.name]
                        bpy.ops.object.select_grouped(type='PARENT')
                else:
                    # if self.set_profile:
                    self.curve_obj.active_material = bpy.data.materials[existing_mat.name]
                    # else:
                    self.mesh_obj.active_material = bpy.data.materials[existing_mat.name]

        # MOVE MESH TO COLLECTION
        # if addon_pref.add_to_collection:
        #     for area in bpy.context.screen.areas:
        #         if area.type == 'VIEW_3D':
        #             if not area.spaces.active.local_view:
        #                 self.move_active_object_to_collection(context)

        if proflow.draw:
            bpy.ops.object.mode_set(mode='OBJECT')
            # Hide start et end cap
            if context.object.type == 'MESH':
                for mod in context.object.modifiers:
                    if mod.type == 'ARRAY':
                        if mod.start_cap is not None:
                            self.array_start_cap = mod.start_cap
                            self.array_start_cap.hide_set(True)
                        if mod.end_cap is not None:
                            self.array_end_cap = mod.end_cap
                            self.array_end_cap.hide_set(True)

            bpy.ops.object.select_all(action='DESELECT')

            context.view_layer.objects.active = self.curve_obj
            self.curve_obj.select_set(state=True)

            # MOVE CURVE TO COLLECTION
            if addon_pref.add_to_collection:
                for area in bpy.context.screen.areas:
                    if area.type == 'VIEW_3D':
                        if not area.spaces.active.local_view:
                            self.move_active_object_to_collection(context)

            bpy.ops.object.mode_set(mode='EDIT')
            if proflow.draw_on_surface:
                context.scene.tool_settings.curve_paint_settings.depth_mode = 'SURFACE'
            else:
                context.scene.tool_settings.curve_paint_settings.depth_mode = 'CURSOR'
            bpy.ops.curve.draw('INVOKE_DEFAULT', error_threshold=0.15, fit_method='REFIT', corner_angle=70,
                               use_cyclic=False, stroke=[],
                               wait_for_input=True)

            bpy.ops.wm.tool_set_by_id(name="builtin.select")



        else:
            if proflow.edit_with_modal:
                bpy.ops.object.proflow_modal_rotate_scale('INVOKE_DEFAULT')

        # Restore active collection
        layer_collection = context.view_layer.layer_collection
        layerColl = recurLayerCollection(layer_collection, proflow.active_coll)
        context.view_layer.active_layer_collection = layerColl
        # if addon_pref.add_to_collection:
        #     for area in bpy.context.screen.areas:
        #         if area.type == 'VIEW_3D':
        #             if area.spaces.active.local_view:
        #                 if proflow.active_coll.strip() != "":
        #                     layer_collection = context.view_layer.layer_collection
        #                     layerColl = recurLayerCollection(layer_collection, proflow.active_coll)
        #                     context.view_layer.active_layer_collection = layerColl

        # print("CURVE: ", self.curve_obj.name)
        # print("ACTIVE OBJECT: ", context.active_object.name)

        # # Parent assets to Active Object
        # if proflow.parent_asset and proflow.active_mesh:
        #     self.curve_obj.select_set(state=True)
        #     self.mesh_obj.select_set(state=True)
        #     context.view_layer.objects.active = proflow.active_mesh
        #     proflow.active_mesh.select_set(state=True)
        #     bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)
        #
        #     bpy.ops.object.select_all(action='DESELECT')
        #
        #     context.view_layer.objects.active = proflow.active_mesh
        #     proflow.active_mesh.select_set(state=True)

        # ------------------------------------------
        # FINALIZE
        # ------------------------------------------
        if proflow.is_set_replace_asset:
            proflow.replace_asset = True
        if proflow.is_set_mat:
            addon_pref.copy_material = True

        context.object.show_in_front = False
        if self.curve_obj:
            self.curve_obj = None
        if self.mesh_obj:
            self.mesh_obj = None
        # proflow.on_curve = False
        proflow.active_mesh = None
        proflow.auxiliary_mesh = None
        proflow.active_coll = ""
        # proflow.use_plane_instance = False
        proflow.draw = False
        proflow.no_material = False
        proflow.from_booleans = False
        proflow.act_obj_material_name = ""
        if proflow.use_selected_custom_mesh:
            proflow.use_custom_mesh = False
            proflow.custom_mesh = None
        proflow.use_selected_custom_mesh = False


        context.area.tag_redraw()



        return {'FINISHED'}

def Union_Objects(context):
    act_obj = context.active_object
    act_obj.select_set(state=False)

    context.view_layer.objects.active = context.selected_objects[0]

    for selObj in context.selected_objects:
        if selObj != context.active_object and (selObj.type == "MESH"):
            obj = context.active_object
            newMod = obj.modifiers.new("Boolean" + selObj.name, "BOOLEAN")
            newMod.operation = 'UNION'
            newMod.object = selObj
            bpy.ops.object.modifier_apply(modifier=newMod.name)
            bpy.ops.object.select_all(action='DESELECT')
            selObj.select_set(state=True)
            bpy.ops.object.delete()
            obj.select_set(state=True)

    context.view_layer.objects.active = act_obj
    act_obj.select_set(state=True)

class PROFLOW_OT_delete_floater_and_curve(Operator):
    '''
    DELETE FLOATER AND CURVE

    This will delete the FLOATER and the Curve
    '''
    bl_idname = "object.proflow_delete_floater_and_curve"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        for obj in context.selected_objects:
            if obj.type == 'CURVE':
                obj.select_set(state=False)

        for obj in context.selected_objects:
            if obj.type == 'MESH':
                if obj.children:
                    for child in obj.children:
                        bpy.data.objects.remove(child, do_unlink=True)
                if obj.modifiers:
                    for mod in obj.modifiers:
                        if mod.type == 'ARRAY':
                            if mod.start_cap is not None:
                                array_start_cap = mod.start_cap
                                bpy.data.objects.remove(array_start_cap, do_unlink=True)

                            if mod.end_cap is not None:
                                array_end_cap = mod.end_cap
                                bpy.data.objects.remove(array_end_cap, do_unlink=True)

                        if mod.type == 'CURVE':
                            curve = mod.object
                            bpy.data.objects.remove(obj, do_unlink=True)
                            bpy.data.objects.remove(curve, do_unlink=True)

        return {'FINISHED'}

def Show_Hide_Modifiers(context):

    bpy.ops.object.show_hide_skrink_data()

class PROFLOW_OT_show_hide_skrink_data(Operator):
    '''
    ENABLE/DISABLE MODIFIERS

    This tool will Disable some modifiers that can slow down the edition of you object

    CLICK - ENABLE/DISABLE on the selection
    CTRL + CLICK - DISABLE on all the objects in the scene
    SHIFT + CLICK - ENABLE on all the objects in the scene
    '''
    bl_idname = "object.show_hide_skrink_data"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def invoke(self, context, event):
        act_obj = context.active_object

        sel = []

        for obj in context.scene.objects:
            if obj.modifiers:
                for mod in obj.modifiers:
                    # if event.alt:
                    if event.ctrl or event.shift:
                        if mod.type == 'SHRINKWRAP':
                            sel.append(obj)
                    else:
                        if mod.type == 'SHRINKWRAP' and mod.target == act_obj:
                            sel.append(obj)
                        elif mod.type == 'SHRINKWRAP' and mod.auxiliary_target == act_obj:
                            sel.append(obj)

        # subsurf_off = False
        # shrinkwrap_visible = False
        # subsurf_invisible = False
        for obj in sel:
            for mod in obj.modifiers:
                # if mod.type == 'SHRINKWRAP' :
                #     if mod.show_viewport == True:
                #         shrinkwrap_visible = True
                #
                # if mod.type == 'SUBSURF' :
                #     if mod.show_viewport == False:
                #         subsurf_invisible = True
                #
                # if shrinkwrap_visible and subsurf_invisible:
                #     subsurf_off = True

                if mod.type in {'DATA_TRANSFER', 'SHRINKWRAP','SUBSURF'}:
                    if event.ctrl:
                        mod.show_viewport = False
                        self.report({'INFO'}, "Modifiers are Invisible")
                    elif event.shift:
                        mod.show_viewport = True
                        self.report({'INFO'}, "Modifiers are Visible")

                    else:
                        if mod.show_viewport:
                            mod.show_viewport = False
                            self.report({'INFO'}, "Modifiers are Invisible")
                        else:
                            mod.show_viewport = True
                            self.report({'INFO'}, "Modifiers are Visible")

                # if subsurf_off == False:
                #     if mod.type == 'SUBSURF':
                #         mod.show_viewport = not mod.show_viewport

        del sel[:]


        return {'FINISHED'}

class PROFLOW_OT_select_floater_and_curve(Operator):
    '''
    SELECT FLOATER AND CURVE

    it will select both Floater and Curve

    CLICK - Select Asset and Curve
    SHIFT + CLICK - Select Only Curve or only Instance
    '''
    bl_idname = "object.proflow_select_floater_and_curve"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def invoke(self, context, event):
        act_obj = context.active_object

        for obj in context.selected_objects:
            if event.shift:
                if obj.name.startswith('proflow_plane_instance'):
                    bpy.ops.object.select_grouped(type='CHILDREN')

                else:
                    if obj.modifiers:
                        for mod in obj.modifiers:
                            if mod.type == 'CURVE':
                                curve = mod.object
                                curve.select_set(state=True)
                                obj.select_set(state=False)

                for obj in context.selected_objects:
                    context.view_layer.objects.active = obj

                return {'FINISHED'}
            else:
                if obj.type == 'MESH':
                    if obj.modifiers:
                        for mod in obj.modifiers:
                            if mod.type == 'SUBSURF':
                                mod.show_viewport = False
                            if mod.type == 'CURVE':
                                curve = mod.object
                                curve.select_set(state=True)


            bpy.ops.object.select_grouped(extend=True, type='CHILDREN_RECURSIVE')

        return {'FINISHED'}

class PROFLOW_OT_select_curve_profile(Operator):
    '''
    SELECT CURVE PROFILE

    it will select Prfile used by the Curve
    '''
    bl_idname = "object.proflow_select_curve_profile"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        self.current_curve = context.object

        if context.object.data.bevel_object is not None:
            self.profile = context.object.data.bevel_object

        bpy.ops.object.select_all(action='DESELECT')

        self.profile.select_set(state=True)
        context.view_layer.objects.active = self.profile

        return {'FINISHED'}

class PROFLOW_OT_select_curve_master(Operator):
    '''
    SELECT CURVE master

    it will select Profile used by the Curve
    '''
    bl_idname = "object.proflow_select_curve_master"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        self.profile = context.object

        for obj in context.scene.objects:
            if obj.type == 'CURVE':
                if obj.data.bevel_object == context.object:
                    self.is_curve_master = obj

        bpy.ops.object.select_all(action='DESELECT')
        self.is_curve_master.select_set(state=True)
        context.view_layer.objects.active = self.is_curve_master

        return {'FINISHED'}

#Parent Unparent
def find_parent(ob, lst):

    lst.append(ob)
    if ob.parent:
        find_parent(ob.parent, lst)

# select_hierarchy
def find_child(ob, lst):
    for x in ob.children:
        lst.append(x)
        find_child(x, lst)

def Last_Parent():
    lst = []  # list for all objects
    find_parent(bpy.context.object, lst)  # find parent
    if len(lst) > 0:
        last_parent = lst[0]
        for p in lst:
            # parent of everything has no parent itself
            if not p.parent: last_parent = p
            bpy.ops.object.select_all(action='DESELECT')
            bpy.context.view_layer.objects.active = last_parent
            last_parent.select_set(state=True)

class PROFLOW_OT_set_custom_mesh_from_sel(Operator):
    '''
    SET SELECTION AS CUSTOM MESH

    it will set the selection to use it as custom Mesh
    '''
    bl_idname = "object.proflow_set_custom_mesh_from_sel"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        proflow = context.window_manager.proflow
        if len(context.selected_objects) == 1:
            for o in context.selected_objects:
                if o.type in {'MESH', 'FONT', 'CURVE', 'EMPTY'}:
                    proflow.custom_mesh = o

        return {'FINISHED'}

class PROFLOW_OT_remove_profile(Operator):
    '''
    DELETE CURVE & PROFILE

    it will select Profile used by the Curve

    CLICK - Delete Curve and Profile
    SHIFT - Remove Profile
    '''
    bl_idname = "object.proflow_remove_profile"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def invoke(self, context, event):
        self.curve = context.object

        if event.shift:
            self.curve.data.bevel_object = None
            bpy.context.object.data.bevel_depth = 0
        else:
            profile = self.curve.data.bevel_object
            bpy.data.objects.remove(profile, do_unlink=True)
            bpy.data.objects.remove(self.curve, do_unlink=True)


        return {'FINISHED'}

class PROFLOW_OT_select_hierarchy(bpy.types.Operator):
    bl_idname = "object.proflow_select_hierarchy"
    bl_label = "select hierarchy"
    bl_options = {"UNDO"}

    def execute(self, context):
        lst = []  # list for all objects
        find_parent(context.object, lst)  # find parent

        if len(lst) > 0:

            last_parent = lst[0]
            for p in lst:
                # parent of everything has no parent itself
                if not p.parent: last_parent = p

            # find children for each parent
            for c in lst: find_child(c, lst)

            # select all
            for x in lst: x.select_set(state=True)

            # make parent active
            context.view_layer.objects.active = last_parent

        # print(lst)
        # context.space_data.pivot_point = 'ACTIVE_ELEMENT'
        return {'FINISHED'}

def local_rotate(obj, axis, angle):
    obj.matrix_world @= Matrix.Rotation(math.radians(angle), 4, axis)

class PROFLOW_OT_convert_to_mesh(Operator):
    '''
    CONVERT TO MESH

    It will convert the Floater to poly.
    It will apply the modifiers and delete the curve
    '''
    bl_idname = "object.proflow_convert_to_mesh"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        proflow = context.window_manager.proflow
        instance = False

        for obj in context.selected_objects:
            # si on a une courbe avec un profile on converti et on quitte et on supprime le profile
            if obj.type == 'CURVE':
                if obj.data.bevel_object is not None:
                    profile = obj.data.bevel_object
                    bpy.ops.object.convert(target='MESH')
                    bpy.data.objects.remove(profile, do_unlink=True)
                    return {'FINISHED'}

            # sinon, on set la courbe pour la supprimer plus tard
            elif obj.type in {'MESH', 'FONT'}:
                for mod in obj.modifiers:
                    if mod.type == 'CURVE':
                        self.curve = mod.object

                # si l'objet est de base on le converti et on supprime la courbe
                if not obj.name.startswith("proflow_plane_instance"):
                    if proflow.remove_subsurf:
                        for mod in obj.modifiers:
                            if mod.type == 'SUBSURF':
                                bpy.ops.object.modifier_remove(modifier=mod.name)
                    bpy.ops.object.convert(target='MESH')
                    bpy.data.objects.remove(self.curve, do_unlink=True)
                    return {'FINISHED'}

                # si plan avec instance
                elif obj.name.startswith("proflow_plane_instance"):
                    self.plane_mesh = obj

                    if obj.children:
                        instance = True
                        for child in obj.children:
                            self.instance_obj = child
                        bpy.ops.object.duplicates_make_real(use_base_parent=True, use_hierarchy=True)
                        if proflow.remove_subsurf:
                            for mod in obj.modifiers:
                                if mod.type == 'SUBSURF':
                                    bpy.ops.object.modifier_remove(modifier=mod.name)

                        bpy.data.objects.remove(self.curve, do_unlink=True)


        if instance:
            # on supprime l'objet de base
            bpy.data.objects.remove(self.instance_obj, do_unlink=True)
            self.plane_mesh.select_set(state=True)
            context.view_layer.objects.active = self.plane_mesh
            bpy.ops.object.proflow_select_hierarchy()
            self.plane_mesh.select_set(state=False)

            has_empty = False
            for obj in context.selected_objects:
                if obj.type == 'EMPTY':
                    has_empty = True
                    # self.empty = obj

            if has_empty:
                self.empty = [o for o in context.selected_objects if o.type == 'EMPTY']
            self.sel = [o for o in context.selected_objects if o.type == 'MESH']

            bpy.ops.object.select_all(action='DESELECT')

            for obj in self.sel:
                obj.select_set(state=True)
                context.view_layer.objects.active = obj
                if proflow.remove_subsurf:
                    if obj.type == 'MESH':
                        for mod in obj.modifiers:
                            if mod.type == 'SUBSURF':
                                bpy.ops.object.modifier_remove(modifier=mod.name)

                bpy.ops.object.make_single_user(type='SELECTED_OBJECTS', object=True, obdata=True, material=False, animation=False)
                bpy.ops.object.convert(target='MESH')

                # bpy.ops.object.join()
                self.converted_mesh = context.active_object
                bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')

                # # Random Rotation
                # if proflow.random_location:
                #     max_z = 1
                #     loc_x = random.randint(max_z * -1, max_z)
                #     loc_y = random.randint(max_z * -1, max_z)
                #     loc_z = 0
                #     curent_loc_x = self.converted_mesh.location[0]
                #     curent_loc_y = self.converted_mesh.location[1]
                #     curent_loc_z = self.converted_mesh.location[2]
                #     self.converted_mesh.location[:] = (curent_loc_x + loc_x * 0.05, curent_loc_y + loc_y * 0.05, curent_loc_z + loc_z * 0.05)
                #
                # # Random Rotation
                # if proflow.random_rotation:
                #     max_z = 5000
                #     my_z = random.randint(max_z * -1, max_z)
                #     ro_z = math.radians(my_z)
                #     local_rotate(obj, 'Z', ro_z)
                #
                # # Random Scale
                # if proflow.random_scale:
                #     # random scale
                #     min_scale = 0.9
                #     max_scale = 1.2
                #     scale_factor = 0.1
                #
                #     for obj in bpy.context.selected_objects:
                #         scale = uniform(min_scale, max_scale) * Vector((1, 1, 1))
                #         obj.scale = scale



                if proflow.join_meshes:
                    bpy.ops.object.join()
                    bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')
                    bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)

            if has_empty:
                for obj in self.empty:
                    obj.select_set(state=True)
                    context.view_layer.objects.active = obj
                    bpy.data.objects.remove(obj, do_unlink=True)
            #
            bpy.data.objects.remove(self.plane_mesh, do_unlink=True)

            context.view_layer.objects.active = self.converted_mesh

            if proflow.link_objects:
                bpy.ops.object.make_links_data(type='OBDATA')

            if not proflow.join_meshes and proflow.random_transforms:
                bpy.ops.object.proflow_loc_rot_scale_random('INVOKE_DEFAULT')

        return {'FINISHED'}

class PROFLOW_OT_select_curve(Operator):
    '''
    SELECT THE CURVE AND EDIT IT

    You can Change the Resolution, Direction,
    the Spline type (Poly or Bezier),
    the handle type (Auto, Free...),
    Etc

    CLICK - Select the Curve and go in EDIT MODE
    SHIFT + CLICK - Select the Curve

    '''
    bl_idname = "object.proflow_select_curve"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def invoke(self, context, event):
        act_obj = context.active_object

        for obj in context.selected_objects:
            if obj.modifiers:
                for mod in obj.modifiers:
                    if mod.type == 'CURVE':
                        self.curve = mod.object
                        bpy.ops.object.select_all(action='DESELECT')
                        self.curve.select_set(state=True)
                        context.view_layer.objects.active = self.curve
                        if self.curve.mode == 'EDIT':
                            bpy.ops.object.mode_set(mode='OBJECT')
                            bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')

                            bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.view3d.snap_cursor_to_selected()

        self.curve.select_set(state=False)
        context.view_layer.objects.active = act_obj
        act_obj.select_set(state=True)
        bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)
        act_obj.select_set(state=False)
        context.view_layer.objects.active = self.curve
        self.curve.select_set(state=True)
        if event.shift:
            bpy.ops.object.mode_set(mode='OBJECT')
        else:
            bpy.ops.object.mode_set(mode='EDIT')

        return {'FINISHED'}

class PROFLOW_OT_select_flow(Operator):
    '''
    SELECT THE Floater OBJECT AND EDIT IT

    You can Change the form of the Floater with modifiers,
    You can bend the Floater, change the Array Space, etc
    '''
    bl_idname = "object.proflow_select_floater"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        self.curve = context.active_object
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='MEDIAN')
        bpy.ops.view3d.snap_cursor_to_selected()

        for obj in context.scene.objects:
            for mod in obj.modifiers:
                if mod.type == 'CURVE' and mod.object == bpy.data.objects[self.curve.name]:
                    # print("MESH: ", obj.name)
                    self.floater = obj
                    self.curve.select_set(state=False)
                    context.view_layer.objects.active = obj
                    obj.select_set(state=True)
                    bpy.ops.view3d.snap_selected_to_cursor(use_offset=False)


        return {'FINISHED'}




class PROFLOW_OT_edit_curve_from_mesh(Operator):
    '''
    CURVE SETTINGS
    '''
    bl_idname = "object.edit_curve_from_mesh"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    edit_curve_modes: EnumProperty(
        items=(('bezier', "Bezier", ""),
               ('poly', "Poly", ""),
               ('cyclic', "Cyclic", ""),
               ),
        default='bezier')

    def execute(self, context):

        self.mesh = context.active_object

        # Set Curve
        if self.mesh.modifiers:
            for mod in self.mesh.modifiers:
                if mod.type == 'CURVE' and mod.name == 'PFLOW - Curve':
                    self.curve = mod.object

        bpy.ops.object.select_all(action='DESELECT')
        context.view_layer.objects.active = self.curve
        self.curve.select_set(state=True)
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')

        if self.edit_curve_modes == 'bezier':
            bpy.ops.curve.spline_type_set(type='BEZIER')
            bpy.ops.curve.handle_type_set(type='AUTOMATIC')
            self.report({'INFO'}, "Cyclic BEZIER")

        elif self.edit_curve_modes == 'poly':
            bpy.ops.curve.spline_type_set(type='POLY')
            self.report({'INFO'}, "Curve in POLY")

        elif self.edit_curve_modes == 'cyclic':
            act_spline = context.object.data.splines.active
            if act_spline is not None:
                if act_spline.type == 'BEZIER':
                    if act_spline.use_cyclic_u == False:
                        act_spline.use_cyclic_u = True
                        self.report({'INFO'}, "Cyclic ON")
                    else:
                        act_spline.use_cyclic_u = False
                        self.report({'INFO'}, "Cyclic OFF")
                else:
                    self.report({'INFO'}, "Curve in POLY")


        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.select_all(action='DESELECT')
        context.view_layer.objects.active = self.mesh
        self.mesh.select_set(state=True)

        return {'FINISHED'}


class PROFLOW_OT_curve_modes(Operator):
    '''
    CURVE SETTINGS
    '''
    bl_idname = "object.proflow_delete_curve_mode"
    bl_label = ""
    bl_options = {'REGISTER', 'UNDO'}

    curve_modes: EnumProperty(
        items=(('bezier', "Bezier", ""),
               ('poly', "Poly", ""),
               ('auto', "Auto", ""),
               ('vector', "Vector", ""),
               ('free', "Free", ""),
               ('align', "Align", ""),
               ('switch_direction', "Switch Direction", ""),
               ('recalculate_normals', "Recalculate Normals", ""),
               ('cyclic', "Cyclic", ""),
               ),
        default='bezier')

    def execute(self, context):
        act_obj = context.active_object
        edit_mode = False
        if act_obj.mode == 'EDIT':
            edit_mode = True

        if act_obj.type == 'CURVE':

            if not self.curve_modes == 'cyclic':

                if not edit_mode:
                    bpy.ops.object.mode_set(mode='EDIT')

                if self.curve_modes == 'bezier':
                    bpy.ops.curve.spline_type_set(type='BEZIER')
                    bpy.ops.curve.handle_type_set(type='AUTOMATIC')
                    self.report({'INFO'}, "Curve in BEZIER")
                elif self.curve_modes == 'poly':
                    bpy.ops.curve.spline_type_set(type='POLY')
                    self.report({'INFO'}, "Curve in POLY")

                elif self.curve_modes == 'auto':
                    bpy.ops.curve.handle_type_set(type='AUTOMATIC')
                elif self.curve_modes == 'vector':
                    bpy.ops.curve.handle_type_set(type='VECTOR')
                elif self.curve_modes == 'free':
                    bpy.ops.curve.handle_type_set(type='FREE_ALIGN')
                elif self.curve_modes == 'align':
                    bpy.ops.curve.handle_type_set(type='ALIGNED')
                elif self.curve_modes == 'switch_direction':
                    bpy.ops.curve.switch_direction()
                elif self.curve_modes == 'recalculate_normals':
                    bpy.ops.curve.normals_make_consistent()

                if not edit_mode:
                    bpy.ops.object.mode_set(mode='OBJECT')

            else:
                act_obj = context.active_object
                act_spline = act_obj.data.splines.active
                if act_spline.use_cyclic_u == False:
                    act_spline.use_cyclic_u = True
                else:
                    act_spline.use_cyclic_u = False

        return {'FINISHED'}

# -----------------------------------------------------
# REGISTER/UNREGISTER
# -----------------------------------------------------
CLASSES =  [PROFLOW_OT_from_boolean,
            PROFLOW_OT_finalize,
            PROFLOW_OT_flow_loop,
            PROFLOW_OT_draw_flow,
            PROFLOW_OT_select_curve,
            PROFLOW_OT_select_flow,
            PROFLOW_OT_convert_to_mesh,
            PROFLOW_OT_delete_floater_and_curve,
            PROFLOW_OT_select_floater_and_curve,
            PROFLOW_OT_replace_asset,
            PROFLOW_OT_rot_increments,
            PROFLOW_OT_curve_modes,
            PROFLOW_OT_edit_curve_from_mesh,
            PROFLOW_OT_select_curve_profile,
            PROFLOW_OT_select_curve_master,
            PROFLOW_OT_remove_profile,
            PROFLOW_OT_select_hierarchy,
            PROFLOW_OT_set_custom_mesh_from_sel,
            PROFLOW_OT_show_hide_skrink_data
            ]

def register():
    for cls in CLASSES:
        try:
            bpy.utils.register_class(cls)
        except:
            print(f"{cls.__name__} already registred")


def unregister():

    for cls in reversed(CLASSES):
        try:
            bpy.utils.unregister_class(cls)
        except RuntimeError:
            pass
    # for cls in CLASSES:
    #     if hasattr(bpy.types, cls.__name__):
    #         bpy.utils.unregister_class(cls)