# -*- coding:utf-8 -*-

# ProFlow Add-on
# Copyright (C) 2018 Cedric Lepiller aka Pitiwazou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# <pep8 compliant>

import bpy
from bpy.types import Panel, Operator, PropertyGroup
from bpy.props import (StringProperty,
                       BoolProperty,
                       FloatVectorProperty,
                       FloatProperty,
                       EnumProperty,
                       IntProperty)
from .operators import *
from .preview import *
from bpy.types import WindowManager
from .functions import get_addon_preferences
from .custom_icons.icons import load_icons

def PROFLOW_UI(self, context):
    icons = load_icons()
    scn = context.scene
    layout = self.layout
    wm = context.window_manager
    proflow = context.window_manager.proflow
    addon_pref = get_addon_preferences()
    preview_scale = addon_pref.preview_scale
    preview_scale_popup = addon_pref.preview_scale_popup
    preview_show_labels = addon_pref.preview_show_labels

    # addon_pref = get_addon_preferences()
    # add_to_collection = addon_pref.add_to_collection

    # Check props
    if proflow.replace_asset:
        proflow.is_set_replace_asset = True
    else:
        proflow.is_set_replace_asset = False
    if addon_pref.copy_material:
        proflow.is_set_mat = True
    else:
        proflow.is_set_mat = False

    # Check Curve
    self.has_bevel = False
    self.curve_master = False

    # proflow.active_mesh = context.active_object

    if context.object is not None and context.selected_objects and context.object.type == 'CURVE':
        if len(context.selected_objects) == 1:
            self.act_spline = context.object.data.splines.active

            if context.object.data.bevel_object is not None:
                self.has_bevel = True
                self.current_curve = context.object.data
                self.profile = context.object.data.bevel_object
            else:
                self.has_bevel = False

            for obj in context.scene.objects:
                if obj.type == 'CURVE':
                    if obj.data.bevel_object == context.object:
                        self.curve_master = True
                        self.is_curve_master = obj
                    else:
                        self.curve_master = False

    # if len(addon_pref.categories):
    row = layout.row(align=True)
    row.scale_y = 1.2
    row.prop(proflow, "ui_categories", text="")

    row = layout.row(align=True)
    col = row.column()
    col.template_icon_view(wm, "proflow_previews", show_labels=preview_show_labels, scale=preview_scale,
                           scale_popup=preview_scale_popup)

    col = row.column(align=True)
    col.prop(proflow, "use_custom_mesh", text="", icon='MONKEY')
    # col.prop(addon_pref, "copy_material", text="", icon='NODE_MATERIAL')
    # col.prop(addon_pref, "add_to_collection", text="", icon='GROUP')
    # col.prop(proflow, "draw_on_surface", text="", icon='SNAP_NORMAL')
    # col.prop(proflow, "edit_with_modal", text="", icon='MOUSE_MMB')
    col.prop(proflow, "replace_asset", text="", icon='FULLSCREEN_EXIT')
    col.prop(proflow, "use_plane_instance", text="", icon='SNAP_FACE_CENTER')
    # col.prop(proflow, "make_booleans", text="", icon='MOD_BOOLEAN')
    if len(context.selected_objects) == 1:
        # col.separator()
        col.operator("proflow.setup_asset", text="", icon='TOOL_SETTINGS')
    # col.separator()
    col.prop(proflow, "show_options", text="", icon='PREFERENCES')
    # col.operator("proflow.options_popup", text="", icon='PREFERENCES')
    col.menu("PROFLOW_MT_help", text="", icon='QUESTION')


    show_simplify = False
    for obj in context.visible_objects:
        if obj.modifiers:
            for mod in obj.modifiers:
                if mod.type == 'DATA_TRANSFER' and mod.object == context.active_object:
                    if mod.type == 'DATA_TRANSFER':
                        show_simplify = True
                        if mod.show_viewport:
                            proflow.proflow_simplify = True
                        else:
                            proflow.proflow_simplify = False

    if show_simplify :
        if proflow.proflow_simplify:
            col.operator("object.show_hide_skrink_data", icon='RADIOBUT_ON')
        else:
            col.operator("object.show_hide_skrink_data", icon='RADIOBUT_OFF')

    if proflow.show_options:
        box = layout.box()
        split = box.split()
        col = split.column()
        row = col.row(align=True)
        row.label(text="BASIC OPTIONS")
        row = col.row(align=True)
        row.scale_x = 2
        row.scale_y = 1.1
        row.prop(addon_pref, "copy_material", text="", icon='NODE_MATERIAL')
        row.prop(addon_pref, "add_to_collection", text="", icon='GROUP')
        row.prop(proflow, "draw_on_surface", text="", icon='SNAP_NORMAL')
        row.prop(proflow, "edit_with_modal", text="", icon='MOUSE_MMB')
        row.prop(proflow, "keep_wireframe", text="", icon='MOD_WIREFRAME')

        # row.prop(proflow, "Transfert_normals", text="", icon='MOD_DATA_TRANSFER')

        curve = False
        if context.object.modifiers:
            for mod in context.object.modifiers:
                if mod.type == 'CURVE':
                    if mod.name == 'PFLOW - Curve':
                        curve = True

        if curve:
            is_instance = False
            is_empty = False
            if context.object.name.startswith("proflow_plane_instance"):
                is_instance = True
                if context.object.children:
                    for child in context.object.children:
                        if child.type == 'EMPTY':
                            is_empty = True

            row = col.row(align=True)
            row.label(text="CONVERTION OPTIONS")

            row = col.row(align=True)
            row.scale_y = 1.2
            row.prop(proflow, "remove_subsurf", text="Remove Subsurf Modifiers", icon='MOD_SUBSURF')


            if is_instance:
                row = col.row(align=True)
                row.scale_y = 1.2
                row.prop(proflow, "join_meshes", text="Join Meshes", icon='AUTOMERGE_ON')
                if not proflow.join_meshes :
                    row = col.row(align=True)
                    row.scale_y = 1.2
                    row.prop(proflow, "link_objects", text="Link Objects", icon='LINKED')

                    if not is_empty:
                        row = col.row(align=True)
                        row.scale_y = 1.2
                        row.prop(proflow, "random_transforms", text="Random Transforms", icon='STICKY_UVS_DISABLE')




    if proflow.use_custom_mesh:
        row = layout.row(align=True)
        if len(context.selected_objects) == 1 :
            # if proflow.custom_mesh is not None:
            if context.object != proflow.custom_mesh :
                row.operator("object.proflow_set_custom_mesh_from_sel",text="", icon='FILE_REFRESH')

        row.prop_search(proflow, "custom_mesh", scn, "objects", text="")

    # row = layout.row(align=True)

    screw = False
    if context.object is not None and context.selected_objects and context.object.type == 'MESH':
        if context.object.modifiers:
            for mod in context.object.modifiers:
                if mod.type == 'SCREW':
                    screw = True

    if context.object is None or not context.selected_objects:
        box = layout.box()
        split = box.split()
        col = split.column()
        row = col.row(align=True)
        proflow.cyclic = False
        row.scale_x = 2
        row.scale_y = 1.2
        icon = icons.get("icon_draw")
        # row.operator("object.proflow_draw", text='Draw Floater', icon_value=icon.icon_id)
        row.operator("object.proflow_draw", text='Draw Floater', icon='GREASEPENCIL')

    elif len(context.selected_objects) == 1:
        # Si curve selectionne
        if context.object.type == 'CURVE' and context.object.mode == 'OBJECT' and not self.has_bevel and not self.curve_master:
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_y = 1.2
            icon = icons.get("icon_proflow")
            op = row.operator("object.proflow_finalize", text="Add Floater", icon_value=icon.icon_id)
            op.on_curve = True

        # Si edge only
        elif context.object.type == 'MESH' and context.object.mode == 'OBJECT' and not len(
                context.object.data.polygons) and not screw:
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_x = 2
            row.scale_y = 1.2
            icon = icons.get("icon_proflow")
            row.operator("object.proflow_loop", text="Add Floater", icon_value=icon.icon_id)

        # Si mesh et replace asset
        elif context.object.type in {'MESH', 'FONT', 'CURVE'} and context.object.mode == 'OBJECT' and proflow.replace_asset:
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_x = 2
            row.scale_y = 1.2
            row.operator("object.proflow_replace_asset", text='Replace Asset', icon='FULLSCREEN_EXIT')

        # Si mesh
        elif context.object.type in {'MESH', 'FONT'} and context.object.mode == 'OBJECT':
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            proflow.cyclic = False
            row.scale_x = 2
            row.scale_y = 1.2
            # icon = icons.get("icon_draw")
            # row.operator("object.proflow_draw", text='Draw', icon_value=icon.icon_id)
            # icon = icons.get("icon_cut")
            # row.operator("object.proflow_modal_cut", text='Cut', icon_value=icon.icon_id)
            # icon = icons.get("icon_extract")
            # row.operator("object.proflow_modal_extract", text='Edges', icon_value=icon.icon_id)
            row.operator("object.proflow_draw", text='Draw', icon='GREASEPENCIL')
            row.operator("object.proflow_modal_cut", text='Cut', icon='SCULPTMODE_HLT')
            row.operator("object.proflow_modal_extract", text='Edges', icon='EDGESEL')

        # Si mesh en edit
        elif context.object.mode == 'EDIT' and context.object.type in {'MESH', 'FONT'}:
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_y = 1.2
            icon = icons.get("icon_proflow")
            row.operator("object.proflow_loop", text="Add Floater", icon_value=icon.icon_id)

    elif len(context.selected_objects) >= 2:
        box = layout.box()
        split = box.split()
        col = split.column()
        row = col.row(align=True)
        proflow.cyclic = True
        proflow.replace_asset = False
        row.scale_y = 1.2
        icon = icons.get("icon_proflow")
        row.operator("object.proflow_from_boolean", text="Add Floater", icon_value=icon.icon_id)
        row.scale_x = 2
        # icon = icons.get("icon_draw")
        # row.operator("object.proflow_draw", icon_value=icon.icon_id)
        row.operator("object.proflow_draw", icon='GREASEPENCIL')

    if context.object is not None and context.selected_objects and context.object.type in {'MESH', 'FONT'}:
        curve = False
        array = False
        subsurf = False
        displace_1 = False
        displace_2 = False
        bend = False
        weld = False
        shrinkwrap = False
        data = False
        if context.object.modifiers:
            for mod in context.object.modifiers:
                if mod.type == 'CURVE':
                    if mod.name == 'PFLOW - Curve':
                        self.mod_curve = mod
                        curve = True
                if mod.type == 'ARRAY':
                    if mod.name == 'PFLOW - Array':
                        self.mod_array = mod
                        array = True
                if mod.type == 'SUBSURF':
                    # if mod.name == 'Subdivision':
                    self.mod_subsurf = mod
                    subsurf = True
                if mod.type == 'DISPLACE':
                    if mod.name == 'PFLOW - Displace - 1':
                        self.mod_displace_1 = mod
                        displace_1 = True
                    if mod.name == 'PFLOW - Displace - 2':
                        self.mod_displace_2 = mod
                        displace_2 = True
                if mod.type == 'SIMPLE_DEFORM':
                    if mod.name == "PFLOW - Bend":
                        self.mod_bend = mod
                        bend = True
                if mod.type == 'WELD':
                    if mod.name == "PFLOW - Weld":
                        self.mod_weld = mod
                        weld = True
                if mod.type == 'SHRINKWRAP':
                    if mod.name == "PFLOW - Shrinkwrap":
                        self.mod_shrinkwrap = mod
                        shrinkwrap = True
                if mod.type == 'DATA_TRANSFER':
                    if mod.name == "PFLOW - DataTransfer":
                        self.mod_data = mod
                        data = True

        if curve:
            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_y = 1.2
            row.operator("object.proflow_select_curve", text="Edit Curve", icon='CURVE_DATA')
            # row.scale_x = 1.2
            row.operator("object.edit_curve_from_mesh", text="", icon='SMOOTHCURVE').edit_curve_modes = 'bezier'
            row.operator("object.edit_curve_from_mesh", text="", icon='LINCURVE').edit_curve_modes = 'poly'
            row.operator("object.edit_curve_from_mesh", text="", icon='CURVE_BEZCIRCLE').edit_curve_modes = 'cyclic'

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_y = 1.2
            row.operator("object.proflow_scale", text="Scale")

            row.scale_x = 1.2
            op = row.operator("object.proflow_modal_rotate_scale", text="", icon='MOUSE_MMB')
            op.from_panel = True

            row = col.row(align=True)
            row.scale_y = 1.2

            if proflow.directions == 'x':
                row.prop(context.active_object, "rotation_euler", text="Rotation:", index=0)
            elif proflow.directions == 'y':
                row.prop(context.active_object, "rotation_euler", text="Rotation:", index=1)
            elif proflow.directions == 'z':
                row.prop(context.active_object, "rotation_euler", text="Rotation:", index=2)

            row = col.row(align=True)
            row.scale_y = 1.2
            row.operator("object.proflow_rotation_increments", text="45°").rot_increments = 'quarante_cinq'
            row.operator("object.proflow_rotation_increments", text="-45°").rot_increments = 'moins_quarante_cinq'
            row.operator("object.proflow_rotation_increments", text="90°").rot_increments = 'quatre_vingt_dix'
            row.operator("object.proflow_rotation_increments", text="-90°").rot_increments = 'moins_quatre_vingt_dix'
            row.operator("object.proflow_rotation_increments", text="180°").rot_increments = 'cent_quatre_vingt'

            box = layout.box()
            split = box.split()
            col = split.column()
            row = col.row(align=True)
            row.scale_y = 1.2
            row.scale_x = 1.1
            row.operator("object.proflow_select_floater_and_curve", text="Select", icon='RESTRICT_SELECT_OFF')
            row.scale_x = 1.1
            row.operator("object.proflow_convert_to_mesh", text="Convert", icon='SNAP_VOLUME')
            row.scale_x = 1.1
            row.operator("object.proflow_delete_floater_and_curve", text="", icon='TRASH')

            row = layout.row(align=True)
            row.scale_y = 1.2
            row.prop(proflow, "show_modifiers", text="MODIFIERS",
                     icon='TRIA_UP' if proflow.show_modifiers else 'MODIFIER')
            row.prop(proflow, "directions", text="")


            is_instance = False
            if context.object.name.startswith("proflow_plane_instance"):
                is_instance = True

            if proflow.show_modifiers:
                if array:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Array", icon='MOD_ARRAY')
                    row.scale_x = 2
                    row.prop(self.mod_array, "use_merge_vertices", text="", icon='AUTOMERGE_ON')
                    row.prop(self.mod_array, "show_viewport", text="")
                    # if is_instance:
                    #     row = col.row(align=True)
                    #     row.prop(self.mod_array, "constant_offset_displace", text="X:", index=0)
                    #     row.prop(self.mod_array, "constant_offset_displace", text="Y:", index=1)
                    #     row.prop(self.mod_array, "constant_offset_displace", text="Z:", index=2)
                    # else:
                    row = col.row(align=True)
                    row.prop(self.mod_array, "relative_offset_displace", text="X:", index=0)
                    row.prop(self.mod_array, "relative_offset_displace", text="Y:", index=1)
                    row.prop(self.mod_array, "relative_offset_displace", text="Z:", index=2)
                    if context.object.modifiers["PFLOW - Array"].use_merge_vertices:
                        row = col.row(align=True)
                        row.prop(self.mod_array, "merge_threshold", text="Merge Distance:")

                if displace_2:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Vertical Displace", icon='MOD_DISPLACE')
                    row.scale_x = 2
                    row.prop(self.mod_displace_2, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_displace_2, "strength", text="Strength:")
                    row = col.row(align=True)
                    row.prop(self.mod_displace_2, "direction", text="Axis")

                if displace_1:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Horizontal Displace", icon='MOD_DISPLACE')
                    row.scale_x = 2
                    row.prop(self.mod_displace_1, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_displace_1, "strength", text="Strength:")
                    row = col.row(align=True)
                    row.prop(self.mod_displace_1, "direction", text="Axis")

                if bend:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Bend", icon='MOD_SIMPLEDEFORM')
                    row.scale_x = 2
                    row.prop(self.mod_bend, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_bend, "angle", text="Bend:")
                    row = col.row(align=True)
                    row.prop(self.mod_bend, "deform_axis", text="Axis")

                if weld:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Weld", icon='AUTOMERGE_OFF')
                    row.scale_x = 2
                    row.prop(self.mod_weld, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_weld, "merge_threshold", text="Distance:")

                if subsurf:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Subsurf", icon='MOD_SUBSURF')
                    row.scale_x = 2
                    row.prop(self.mod_subsurf, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_subsurf, "levels", text="Viewport:")

                if shrinkwrap:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Shrinkwrap", icon='MOD_SHRINKWRAP')
                    row.scale_x = 2
                    row.prop(self.mod_shrinkwrap, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_shrinkwrap, "target", text="Target")
                    row = col.row(align=True)
                    row.prop(self.mod_shrinkwrap, "auxiliary_target", text="Auxiliary Target")
                    row = col.row(align=True)
                    row.prop(self.mod_shrinkwrap, "use_negative_direction", text="Use Negative Direction")
                    row = col.row(align=True)
                    row.prop(self.mod_shrinkwrap, "use_positive_direction", text="Use Positive Direction")

                if data:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Data Transfer", icon='MOD_DATA_TRANSFER')
                    row.scale_x = 2
                    row.prop(self.mod_data, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_data, "object", text="Object")

                if curve:
                    box = layout.box()
                    split = box.split()
                    col = split.column()
                    row = col.row(align=True)
                    row.label(text="Curve", icon='MOD_CURVE')
                    row.scale_x = 2
                    row.prop(self.mod_curve, "show_viewport", text="")
                    row = col.row(align=True)
                    row.prop(self.mod_curve, "deform_axis", text="Axis")


    elif context.object is not None and context.selected_objects and context.object.type == 'CURVE':
        box = layout.box()
        split = box.split()
        col = split.column()
        row = col.row(align=True)
        row.scale_y = 1.2
        if not self.has_bevel and not self.curve_master:
            row.operator("object.proflow_select_floater", text="Edit Mesh", icon='GP_MULTIFRAME_EDITING')

        elif not self.has_bevel and self.curve_master:
            row.operator("object.proflow_select_curve_master", text="Edit Curve", icon='CURVE_DATA')

        else:
            row.operator("object.proflow_select_curve_profile", text="Edit Profile", icon='RNDCURVE')
            if self.has_bevel:
                row.scale_x = 1.1
                op = row.operator("object.proflow_modal_rotate_scale", text="", icon='MOUSE_MMB')
                op.from_panel = True
            row.scale_x = 1.1
            row.operator("object.proflow_convert_to_mesh", text="", icon='SNAP_VOLUME')
            row.operator("object.proflow_remove_profile", text="", icon='PANEL_CLOSE')

        # row = layout.row()
        # row.scale_y = 1.2
        # row.prop(proflow, "show_curves_settings", text="CURVE SETTINGS",
        #          icon='TRIA_UP' if proflow.show_curves_settings else 'CURVE_DATA')
        # if proflow.show_curves_settings:
        box = layout.box()
        split = box.split()
        col = split.column()
        row = col.row(align=True)
        row.scale_y = 1.2
        row.operator("object.proflow_delete_curve_mode", text="Cyclic", icon='CURVE_BEZCIRCLE').curve_modes = 'cyclic'
        col.separator()
        row = col.row(align=True)
        row.scale_y = 1.2
        row.operator("object.proflow_delete_curve_mode", text="Bezier", icon='SMOOTHCURVE').curve_modes = 'bezier'
        row.operator("object.proflow_delete_curve_mode", text="Poly", icon='LINCURVE').curve_modes = 'poly'
        row = col.row(align=True)
        row.operator("object.proflow_delete_curve_mode", text="Auto").curve_modes = 'auto'
        row.operator("object.proflow_delete_curve_mode", text="Free").curve_modes = 'free'
        row.operator("object.proflow_delete_curve_mode", text="Vector").curve_modes = 'vector'
        row.operator("object.proflow_delete_curve_mode", text="Align").curve_modes = 'align'


        col.separator()

        if self.act_spline is not None:
            if self.act_spline.type == 'BEZIER':
                row = col.row(align=True)
                row.prop(self.act_spline, "resolution_u", text="Curve Resolution")
                # if has_bevel:
                #     row = col.row(align=True)
                #     row.prop(self.profile, "resolution_u", text="Profile Resolution")

        row = col.row(align=True)
        row.operator("object.proflow_delete_curve_mode", text="Switch Direction",
                     icon='CURVE_PATH').curve_modes = 'switch_direction'
        row = col.row(align=True)
        row.operator("object.proflow_delete_curve_mode", text="Recalculate Normals",
                     icon='MOD_NORMALEDIT').curve_modes = 'recalculate_normals'

        # row = col.row(align=True)
        # row.prop("object.data.use_deform_bounds", text="End")

        if self.has_bevel:
            row = col.row(align=True)
            # row.prop(self.act_spline, "use_smooth", text="Smooth", icon='DOT')
            row.prop(self.current_curve, "use_fill_caps", icon='DOT')
            row = col.row(align=True)
            row.prop(self.current_curve, "bevel_factor_start", text="Bevel Start")
            row = col.row(align=True)
            row.prop(self.current_curve, "bevel_factor_end", text="End")
            row = col.row(align=True)
            spline = context.object.data.splines[0]
            row.prop(spline, "use_smooth", text="Smooth")


        # else:
        #     row = col.row(align=True)
        #     row.prop(self.act_spline, "use_smooth", text="Smooth", icon='DOT')

# PANEL
class PROFLOW_PT_panel(Panel):
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "ProFlow"
    bl_category = "Tools"

    def draw_header(self, context):
        icons = load_icons()
        layout = self.layout
        icon = icons.get("icon_proflow")
        layout.label(text="", icon_value=icon.icon_id)

    def draw(self, context):
        layout = self.layout
        PROFLOW_UI(self, context)

# POPUP
class PROFLOW_OT_popup(bpy.types.Operator):
    bl_idname = "proflow.popup"
    bl_label = "ProFlow Popup"

    def execute(self, context):
        return {'FINISHED'}

    def check(self, context):
        return True

    def draw(self, context):
        layout = self.layout
        PROFLOW_UI(self, context)

    def invoke(self, context, event):
        return context.window_manager.invoke_popup(self, width=180)

# SETUP ASSET
class PROFLOW_PT_setup_asset(bpy.types.Operator):
    bl_idname = "proflow.setup_asset"
    bl_label = "Setup the asset"
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):
        self.action(context)
        return {'FINISHED'}

    def check(self, context):
        return True

    custom_property_name: StringProperty(description="")

    array_axis: EnumProperty(
        items=(('x', "x", ""),
               ('y', "y", ""),
               ('z', 'z', '')),
        default='z')

    bend_axis: EnumProperty(
        items=(('x', "x", ""),
               ('y', "y", ""),
               ('z', 'z', '')),
        default='z')
    array_space: FloatProperty(name="Array Space", default=1, min=-1000, max=1000)
    array_use_merge : BoolProperty(name="Array Merge", default=False)
    subsurf : BoolProperty(default = False)
    subsurf_value: IntProperty(name="Subsurf Iterations", default=2, min=0, max=6)
    segments: IntProperty(name="Segments", default=4, min=1, max=13)
    bend : BoolProperty(default=False)
    bend_value: FloatProperty(name="Bend Angle", default=0, min=-1000, max=1000, subtype = 'ANGLE')
    weld: BoolProperty(default=False)
    data: BoolProperty(default=False)
    plane: BoolProperty(default=False)

    def action(self, context):
        act_obj = context.active_object

        # Clean previous values
        if act_obj.get("proflow_array"):
            del act_obj["proflow_array"]
        if act_obj.get("proflow_bend"):
            del act_obj["proflow_bend"]
        if act_obj.get("proflow_subsurf"):
            del act_obj["proflow_subsurf"]
        if act_obj.get("proflow_weld"):
            del act_obj["proflow_weld"]
        if act_obj.get("proflow_data_transfer"):
            del act_obj["proflow_data_transfer"]
        if act_obj.get("proflow_plane_instance"):
            del act_obj["proflow_plane_instance"]

        act_obj["proflow_array"] = {"axis": self.array_axis, "value": self.array_space, "merge": self.array_use_merge}

        if self.bend:
            act_obj["proflow_bend"] = {"axis": self.bend_axis, "value": self.bend_value}

        if self.weld:
            act_obj["proflow_weld"] = True

        if self.data:
            act_obj["proflow_data_transfer"] = True

        if self.subsurf:
            act_obj["proflow_subsurf"] = self.subsurf_value

        if self.plane:
            act_obj["proflow_plane_instance"] = True

        self.report({'INFO'}, "Asset setting Saved")

    def draw(self, context):

        layout = self.layout

        # ARRAY
        layout.label(text="ARRAY", icon='MOD_ARRAY')
        layout.prop(self, "array_axis", expand = True)
        layout.prop(self, "array_space")
        layout.prop(self, "array_use_merge")

        # SUBSURF
        layout.prop(self, "subsurf", icon='MOD_SUBSURF')
        if self.subsurf:
            layout.prop(self, "subsurf_value")

        # BEND
        layout.prop(self, "bend", icon='MOD_SIMPLEDEFORM')
        if self.bend:
            layout.prop(self, "bend_axis", expand = True)
            layout.prop(self, "bend_value")

        layout.separator()

        # WELD
        layout.prop(self, "weld", icon='AUTOMERGE_OFF')

        # DATA TRANSFERT
        layout.prop(self, "data", icon='MOD_DATA_TRANSFER')

        # PLANE INSTANCE
        layout.prop(self, "plane", icon='SNAP_FACE_CENTER')

    def invoke(self, context, event):
        act_obj = context.active_object
        # if context.object:
        # CHECK ARRAY
        if act_obj.get("proflow_array"):
            self.array_space = act_obj["proflow_array"]["value"]
            self.array_axis = act_obj["proflow_array"]["axis"]
            self.array_use_merge = act_obj["proflow_array"]["merge"]

            # CHECK SUBSURF
        if act_obj.get("proflow_subsurf"):
            self.subsurf = True
            self.subsurf_value = act_obj["proflow_subsurf"]

        # CHECK BEND
        if act_obj.get("proflow_bend"):
            self.bend = True
            self.bend_value = act_obj["proflow_bend"]["value"]
            self.bend_axis = act_obj["proflow_bend"]["axis"]

        # CHECK WELD
        if act_obj.get("proflow_weld"):
            self.weld = True

        # CHECK DATA
        if act_obj.get("proflow_data_transfer"):
            self.data = True

        # CHECK PLANE INSTANCE
        if act_obj.get("proflow_plane_instance"):
            self.plane = True


        # return context.window_manager.invoke_props_dialog(self, width=200, height=100)

        return context.window_manager.invoke_props_dialog(self, width=200)

class PROFLOW_MT_help(bpy.types.Menu):
    bl_idname = "PROFLOW_MT_help"
    bl_label = "PROFLOW HELP"

    def draw(self, context):
        icons = load_icons()
        layout = self.layout
        icon = icons.get("icon_proflow")
        layout.label(text="PROFLOW INFORMATION", icon_value=icon.icon_id)
        layout.label(text="")
        # layout.label(text="How to use the tool:")
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="- ProFlow Workflow",
                        icon_value=icon.icon_id).url = "https://youtu.be/A6crlkhxHDU"
        layout.label(text="- select one or two (intersecting) objects")
        layout.label(text="- Choose the Floater you want to use")
        layout.label(text="- Click on Add Floater")
        layout.label(text="- Edit the settings of the curve if necessary")
        layout.label(text="- Click on 'Edit Mesh' to edit the mesh")
        layout.label(text="- You can Edit or Activate the modifiers to your liking")
        layout.label(text="")
        layout.label(text="You can also select one object and click on Draw Floater,")
        layout.label(text="to draw a Floater on the surface of the selection")
        layout.label(text="")
        layout.label(text="You can use a custom Floater from your scene", icon='MONKEY')
        layout.label(text=" by clicking on the Monkey icon")
        layout.label(text="")
        layout.label(text="You can copy the material of the selection to the Floater", icon='NODE_MATERIAL')
        layout.label(text=" by clicking on the material icon")
        layout.label(text="")
        layout.label(text="You can add the Floater and the curve in a collection", icon='GROUP')
        layout.label(text="by clicking on the Collection icon")
        layout.label(text="")
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Booleans",
                        icon_value=icon.icon_id).url = "https://youtu.be/b2A-gWyhXz4"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="The Cut tool",
                        icon_value=icon.icon_id).url = "https://youtu.be/LZPym7L3Jks"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Extract edge Tool",
                        icon_value=icon.icon_id).url = "https://youtu.be/6CWQ17Z-Bw8"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Replace Asset",
                        icon_value=icon.icon_id).url = "https://youtu.be/Hompqnqd2oY"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Use an Asset from your scene",
                        icon_value=icon.icon_id).url = "https://youtu.be/t-zbBj2HNlw"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Create and Use Collection Instances",
                        icon_value=icon.icon_id).url = "https://youtu.be/f7iS2e7FYro"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Setup your asset",
                        icon_value=icon.icon_id).url = "https://youtu.be/hcNKdfLZHvI"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Create your own Asset",
                        icon_value=icon.icon_id).url = "https://youtu.be/DR0inBrRSfU"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Create Transfer Assets",
                        icon_value=icon.icon_id).url = "https://youtu.be/S1AsasVopKc"
        icon = icons.get("icon_youtube")
        layout.operator("wm.url_open", text="Convert with Random values",
                        icon_value=icon.icon_id).url = "https://youtu.be/f7iS2e7FYro"

        #https://www.youtube.com/watch?v=S1AsasVopKc&feature=youtu.be

# -----------------------------------------------------
# REGISTER/UNREGISTER
# -----------------------------------------------------
CLASSES =  [PROFLOW_PT_panel,
            PROFLOW_MT_help,
            PROFLOW_OT_popup,
            PROFLOW_PT_setup_asset
            ]

# Register
def register():
    for cls in CLASSES:
        try:
            bpy.utils.register_class(cls)
        except:
            print(f"{cls.__name__} already registred")

def unregister():

    for cls in reversed(CLASSES):
        try:
            bpy.utils.unregister_class(cls)
        except RuntimeError:
            pass
    # for cls in CLASSES:
    #     if hasattr(bpy.types, cls.__name__):
    #         bpy.utils.unregister_class(cls)


